import numpy as np
from numba import jit

from submm.KIDs import calibrate

J = np.exp(2j * np.pi / 3)
Jc = 1 / J


@jit(nopython=True)
def cardan(a, b, c, d):
    """
    analytical root finding fast: using numba looks like x10 speed up
    returns only the largest real root
    """
    u = np.empty(2, np.complex128)
    z0 = b / 3 / a
    a2, b2 = a * a, b * b
    p = -b2 / 3 / a2 + c / a
    q = (b / 27 * (2 * b2 / a2 - 9 * c / a) + d) / a
    D = -4 * p * p * p - 27 * q * q
    r = np.sqrt(-D / 27 + 0j)
    one_third = 1 / 3.0
    u = ((-q - r) / 2) ** one_third  # 0.33333333333333333333333
    v = ((-q + r) / 2) ** one_third  # 0.33333333333333333333333
    w = u * v
    w0 = np.abs(w + p / 3)
    w1 = np.abs(w * J + p / 3)
    w2 = np.abs(w * Jc + p / 3)
    if w0 < w1:
        if w2 < w0:
            v *= Jc
    elif w2 < w1:
        v *= Jc
    else:
        v *= J
    roots = np.asarray((u + v - z0, u * J + v * Jc - z0, u * Jc + v * J - z0))
    # print(roots)
    where_real = np.where(np.abs(np.imag(roots)) < 1e-15)
    # if len(where_real)>1: print(len(where_real))
    # print(D)
    if D > 0:
        return np.max(np.real(roots))  # three real roots
    else:
        # one real root get the value that has the smallest imaginary component
        return np.real(roots[np.argsort(np.abs(np.imag(roots)))][0])
    # return np.max(np.real(roots[where_real]))
    # return np.asarray((u+v-z0, u*J+v*Jc-z0,u*Jc+v*J-z0))


def amplitude_normalization(x, z):
    """
    # normalize the amplitude varation requires a gain scan
    #flag frequencies to use in amplitude normaliztion
    """
    index_use = np.where(np.abs(x - np.median(x)) > 100000)  # 100kHz away from resonator
    poly = np.polyfit(x[index_use], np.abs(z[index_use]), 2)
    poly_func = np.poly1d(poly)
    normalized_data = z / poly_func(x) * np.median(np.abs(z[index_use]))
    return normalized_data


def amplitude_normalization_sep(gain_x, gain_z, fine_x, fine_z, stream_x, stream_z):
    """
    # normalize the amplitude varation requires a gain scan
    # uses gain scan to normalize does not use fine scan
    #flag frequencies to use in amplitude normaliztion
    """
    index_use = np.where(np.abs(gain_x - np.median(gain_x)) > 100000)  # 100kHz away from resonator
    poly = np.polyfit(gain_x[index_use], np.abs(gain_z[index_use]), 2)
    poly_func = np.poly1d(poly)
    poly_data = poly_func(gain_x)
    normalized_gain = gain_z / poly_data * np.median(np.abs(gain_z[index_use]))
    normalized_fine = fine_z / poly_func(fine_x) * np.median(np.abs(gain_z[index_use]))
    normalized_stream = stream_z / poly_func(stream_x) * np.median(np.abs(gain_z[index_use]))
    amp_norm_dict = {'normalized_gain': normalized_gain,
                     'normalized_fine': normalized_fine,
                     'normalized_stream': normalized_stream,
                     'poly_data': poly_data}
    return amp_norm_dict


def guess_x0_iq_nonlinear(x, z, verbose=False):
    """
    # this is less robust than guess_x0_iq_nonlinear_sep
    # below. it is recommended to use that instead
    #make sure data is sorted from low to high frequency
    """
    sort_index = np.argsort(x)
    x = x[sort_index]
    z = z[sort_index]
    # extract just fine data
    df = np.abs(x - np.roll(x, 1))
    fine_df = np.min(df[np.where(df != 0)])
    fine_z_index = np.where(df < fine_df * 1.1)
    fine_z = z[fine_z_index]
    fine_x = x[fine_z_index]
    # extract the gain scan
    gain_z_index = np.where(df > fine_df * 1.1)
    gain_z = z[gain_z_index]
    gain_x = x[gain_z_index]
    gain_phase = np.arctan2(np.real(gain_z), np.imag(gain_z))

    # guess f0
    fr_guess_index = np.argmin(np.abs(z))
    # fr_guess = x[fr_guess_index]
    fr_guess_index_fine = np.argmin(np.abs(fine_z))
    # below breaks if there is not a right and left side in the fine scan
    if fr_guess_index_fine == 0:
        fr_guess_index_fine = len(fine_x) // 2
    elif fr_guess_index_fine == (len(fine_x) - 1):
        fr_guess_index_fine = len(fine_x) // 2
    fr_guess = fine_x[fr_guess_index_fine]

    # guess Q
    mag_max = np.max(np.abs(fine_z) ** 2)
    mag_min = np.min(np.abs(fine_z) ** 2)
    mag_3dB = (mag_max + mag_min) / 2.
    half_distance = np.abs(fine_z) ** 2 - mag_3dB
    right = half_distance[fr_guess_index_fine:-1]
    left = half_distance[0:fr_guess_index_fine]
    right_index = np.argmin(np.abs(right)) + fr_guess_index_fine
    left_index = np.argmin(np.abs(left))
    Q_guess_Hz = fine_x[right_index] - fine_x[left_index]
    Q_guess = fr_guess / Q_guess_Hz

    # guess amp
    d = np.max(20 * np.log10(np.abs(z))) - np.min(20 * np.log10(np.abs(z)))
    amp_guess = 0.0037848547850284574 + 0.11096782437821565 * d - 0.0055208783469291173 * d ** 2 + 0.00013900471000261687 * d ** 3 + -1.3994861426891861e-06 * d ** 4  # polynomial fit to amp verus depth

    # guess impedance rotation phi
    phi_guess = 0

    # guess non-linearity parameter
    # might be able to guess this by taking the ratio of the distance between min and max distance between iq points in fine sweep
    a_guess = 0

    # i0 and iq guess
    if np.max(np.abs(fine_z)) == np.max(np.abs(z)):
        # if the resonator has an impedance mismatch rotation that makes the fine greater that the cable delay
        i0_guess = np.real(fine_z[np.argmax(np.abs(fine_z))])
        q0_guess = np.imag(fine_z[np.argmax(np.abs(fine_z))])
    else:
        i0_guess = (np.real(fine_z[0]) + np.real(fine_z[-1])) / 2.
        q0_guess = (np.imag(fine_z[0]) + np.imag(fine_z[-1])) / 2.

    # cable delay guess tau
    # y = mx +b
    # m = (y2 - y1)/(x2-x1)
    # b = y-mx
    if len(gain_z) > 1:  # is there a gain scan?
        m = (gain_phase - np.roll(gain_phase, 1)) / (gain_x - np.roll(gain_x, 1))
        b = gain_phase - m * gain_x
        m_best = np.median(m[~np.isnan(m)])
        tau_guess = m_best / (2 * np.pi)
    else:
        tau_guess = 3 * 10 ** -9

    x0 = [fr_guess, Q_guess, amp_guess, phi_guess, a_guess, i0_guess, q0_guess, tau_guess, fr_guess]

    return x0


def guess_x0_mag_nonlinear(x, z, verbose=False):
    """
    # this is less robust than guess_x0_mag_nonlinear_sep
    #below it is recommended to use that instead
    #make sure data is sorted from low to high frequency
    """
    sort_index = np.argsort(x)
    x = x[sort_index]
    z = z[sort_index]
    # extract just fine data
    # this will probably break if there is no fine scan
    df = np.abs(x - np.roll(x, 1))
    fine_df = np.min(df[np.where(df != 0)])
    fine_z_index = np.where(df < fine_df * 1.1)
    fine_z = z[fine_z_index]
    fine_x = x[fine_z_index]
    # extract the gain scan
    gain_z_index = np.where(df > fine_df * 1.1)
    gain_z = z[gain_z_index]
    gain_x = x[gain_z_index]
    gain_phase = np.arctan2(np.real(gain_z), np.imag(gain_z))

    # guess f0
    fr_guess_index = np.argmin(np.abs(z))
    # fr_guess = x[fr_guess_index]
    fr_guess_index_fine = np.argmin(np.abs(fine_z))
    if fr_guess_index_fine == 0:
        fr_guess_index_fine = len(fine_x) // 2
    elif fr_guess_index_fine == (len(fine_x) - 1):
        fr_guess_index_fine = len(fine_x) // 2
    fr_guess = fine_x[fr_guess_index_fine]

    # guess Q
    mag_max = np.max(np.abs(fine_z) ** 2)
    mag_min = np.min(np.abs(fine_z) ** 2)
    mag_3dB = (mag_max + mag_min) / 2.
    half_distance = np.abs(fine_z) ** 2 - mag_3dB
    right = half_distance[fr_guess_index_fine:-1]
    left = half_distance[0:fr_guess_index_fine]
    right_index = np.argmin(np.abs(right)) + fr_guess_index_fine
    left_index = np.argmin(np.abs(left))
    Q_guess_Hz = fine_x[right_index] - fine_x[left_index]
    Q_guess = fr_guess / Q_guess_Hz

    # guess amp
    d = np.max(20 * np.log10(np.abs(z))) - np.min(20 * np.log10(np.abs(z)))
    amp_guess = 0.0037848547850284574 + 0.11096782437821565 * d - 0.0055208783469291173 * d ** 2 + 0.00013900471000261687 * d ** 3 + -1.3994861426891861e-06 * d ** 4  # polynomial fit to amp verus depth

    # guess impedance rotation phi
    phi_guess = 0

    # guess non-linearity parameter
    # might be able to guess this by using the ratio of the distance between min and max distance between iq points
    #   in fine sweep
    a_guess = 0

    # b0 and b1 guess

    if len(gain_z) > 1:
        xlin = (gain_x - fr_guess) / fr_guess
        b1_guess = (np.abs(gain_z)[-1] ** 2 - np.abs(gain_z)[0] ** 2) / (xlin[-1] - xlin[0])
    else:
        xlin = (fine_x - fr_guess) / fr_guess
        b1_guess = (np.abs(fine_z)[-1] ** 2 - np.abs(fine_z)[0] ** 2) / (xlin[-1] - xlin[0])
    b0_guess = np.median(np.abs(gain_z) ** 2)

    x0 = [fr_guess, Q_guess, amp_guess, phi_guess, a_guess, b0_guess, b1_guess, fr_guess]


    return x0


def guess_x0_iq_nonlinear_sep(fine_x, fine_z, gain_x, gain_z, verbose=False):
    """
    # this is the same as guess_x0_iq_nonlinear except that it takes
    # takes the fine scan and the gain scan as seperate variables
    # this runs into less issues when trying to sort out what part of
    # data is fine and what part is gain for the guessing
    #make sure data is sorted from low to high frequency
    """

    # gain phase
    gain_phase = np.arctan2(np.real(gain_z), np.imag(gain_z))

    # guess f0
    fr_guess_index = np.argmin(np.abs(fine_z))
    # below breaks if there is not a right and left side in the fine scan
    if fr_guess_index == 0:
        fr_guess_index = len(fine_x) // 2
    elif fr_guess_index == (len(fine_x) - 1):
        fr_guess_index = len(fine_x) // 2
    fr_guess = fine_x[fr_guess_index]

    # guess Q
    mag_max = np.max(np.abs(fine_z) ** 2)
    mag_min = np.min(np.abs(fine_z) ** 2)
    mag_3dB = (mag_max + mag_min) / 2.
    half_distance = np.abs(fine_z) ** 2 - mag_3dB
    right = half_distance[fr_guess_index:-1]
    left = half_distance[0:fr_guess_index]
    right_index = np.argmin(np.abs(right)) + fr_guess_index
    left_index = np.argmin(np.abs(left))
    Q_guess_Hz = fine_x[right_index] - fine_x[left_index]
    Q_guess = fr_guess / Q_guess_Hz

    # guess amp
    d = np.max(20 * np.log10(np.abs(gain_z))) - np.min(20 * np.log10(np.abs(fine_z)))
    amp_guess = 0.0037848547850284574 + 0.11096782437821565 * d - 0.0055208783469291173 * d ** 2 + 0.00013900471000261687 * d ** 3 + -1.3994861426891861e-06 * d ** 4  # polynomial fit to amp verus depth

    # guess impedance rotation phi
    # phi_guess = 0
    # guess impedance rotation phi
    # fit a circle to the iq loop
    xc, yc, R, residu = calibrate.leastsq_circle(np.real(fine_z), np.imag(fine_z))
    # compute angle between (off_res,off_res),(0,0) and (off_ress,off_res),(xc,yc) of the the fitted circle
    off_res_i, off_res_q = (np.real(fine_z[0]) + np.real(fine_z[-1])) / 2., (
            np.imag(fine_z[0]) + np.imag(fine_z[-1])) / 2.
    x1, y1, = -off_res_i, -off_res_q
    x2, y2 = xc - off_res_i, yc - off_res_q
    dot = x1 * x2 + y1 * y2  # dot product
    det = x1 * y2 - y1 * x2  # determinant
    angle = np.arctan2(det, dot)
    phi_guess = angle

    # if phi is large better re guess f0
    # f0 should be the farthers from the off res point
    if (np.abs(phi_guess) > 0.3):
        dist1 = np.sqrt((np.real(fine_z[0]) - np.real(fine_z)) ** 2 + (np.imag(fine_z[0]) - np.imag(fine_z)) ** 2)
        dist2 = np.sqrt((np.real(fine_z[-1]) - np.real(fine_z)) ** 2 + (np.imag(fine_z[-1]) - np.imag(fine_z)) ** 2)
        fr_guess_index = np.argmax((dist1 + dist2))
        fr_guess = fine_x[fr_guess_index]
        # also fix the Q gues
        fine_z_derot = (fine_z - (off_res_i + 1.j * off_res_q)) * np.exp(1j * (-phi_guess)) + (
                off_res_i + 1.j * off_res_q)
        # fr_guess_index = np.argmin(np.abs(fine_z_derot))
        # fr_guess = fine_x[fr_guess_index]
        mag_max = np.max(np.abs(fine_z_derot) ** 2)
        mag_min = np.min(np.abs(fine_z_derot) ** 2)
        mag_3dB = (mag_max + mag_min) / 2.
        half_distance = np.abs(fine_z_derot) ** 2 - mag_3dB
        right = half_distance[np.argmin(np.abs(fine_z_derot)):-1]
        left = half_distance[0:np.argmin(np.abs(fine_z_derot))]
        right_index = np.argmin(np.abs(right)) + np.argmin(np.abs(fine_z_derot))
        left_index = np.argmin(np.abs(left))
        Q_guess_Hz = fine_x[right_index] - fine_x[left_index]
        Q_guess = fr_guess / Q_guess_Hz
        # also fix amp guess
        d = np.max(20 * np.log10(np.abs(gain_z))) - np.min(20 * np.log10(np.abs(fine_z_derot)))
        amp_guess = 0.0037848547850284574 + 0.11096782437821565 * d - 0.0055208783469291173 * d ** 2 + 0.00013900471000261687 * d ** 3 + -1.3994861426891861e-06 * d ** 4

    # guess non-linearity parameter
    # might be able to guess this by taking the ratio of the distance between min and max distance between iq points
    # in fine sweep
    a_guess = 0

    # i0 and iq guess
    if np.max(np.abs(fine_z)) > np.max(np.abs(gain_z)):
        # if the resonator has an impedance mismatch rotation that makes the fine greater that the cable delay
        i0_guess = np.real(fine_z[np.argmax(np.abs(fine_z))])
        q0_guess = np.imag(fine_z[np.argmax(np.abs(fine_z))])
    else:
        i0_guess = (np.real(fine_z[0]) + np.real(fine_z[-1])) / 2.
        q0_guess = (np.imag(fine_z[0]) + np.imag(fine_z[-1])) / 2.

    # cable delay guess tau
    # y = mx +b
    # m = (y2 - y1)/(x2-x1)
    # b = y-mx
    m = (gain_phase - np.roll(gain_phase, 1)) / (gain_x - np.roll(gain_x, 1))
    b = gain_phase - m * gain_x
    m_best = np.median(m[~np.isnan(m)])
    tau_guess = m_best / (2 * np.pi)

    if verbose == True:
        print("fr guess  = %.3f MHz" % (fr_guess / 10 ** 6))
        print("Q guess   = %.2f kHz, %.1f" % ((Q_guess_Hz / 10 ** 3), Q_guess))
        print("amp guess = %.2f" % amp_guess)
        print("phi guess = %.2f" % phi_guess)
        print("i0 guess  = %.2f" % i0_guess)
        print("q0 guess  = %.2f" % q0_guess)
        print("tau guess = %.2f x 10^-7" % (tau_guess / 10 ** -7))

    x0 = [fr_guess, Q_guess, amp_guess, phi_guess, a_guess, i0_guess, q0_guess, tau_guess, fr_guess]
    return x0


def guess_x0_mag_nonlinear_sep(fine_x, fine_z, gain_x, gain_z, verbose=False):
    """
    # this is the same as guess_x0_mag_nonlinear except that it takes
    # takes the fine scan and the gain scan as seperate variables
    # this runs into less issues when trying to sort out what part of
    # data is fine and what part is gain for the guessing
    #make sure data is sorted from low to high frequency
    """

    # phase of gain
    gain_phase = np.arctan2(np.real(gain_z), np.imag(gain_z))

    # guess f0
    fr_guess_index = np.argmin(np.abs(fine_z))
    # protect against guessing the first or last data points
    if fr_guess_index == 0:
        fr_guess_index = len(fine_x) // 2
    elif fr_guess_index == (len(fine_x) - 1):
        fr_guess_index = len(fine_x) // 2
    fr_guess = fine_x[fr_guess_index]

    # guess Q
    mag_max = np.max(np.abs(fine_z) ** 2)
    mag_min = np.min(np.abs(fine_z) ** 2)
    mag_3dB = (mag_max + mag_min) / 2.
    half_distance = np.abs(fine_z) ** 2 - mag_3dB
    right = half_distance[fr_guess_index:-1]
    left = half_distance[0:fr_guess_index]
    right_index = np.argmin(np.abs(right)) + fr_guess_index
    left_index = np.argmin(np.abs(left))
    Q_guess_Hz = fine_x[right_index] - fine_x[left_index]
    Q_guess = fr_guess / Q_guess_Hz

    # guess amp
    d = np.max(20 * np.log10(np.abs(gain_z))) - np.min(20 * np.log10(np.abs(fine_z)))
    amp_guess = 0.0037848547850284574 + 0.11096782437821565 * d - 0.0055208783469291173 * d ** 2 + \
                0.00013900471000261687 * d ** 3 + -1.3994861426891861e-06 * d ** 4
    # polynomial fit to amp versus depth calculated empirically

    # guess impedance rotation phi
    # fit a circle to the iq loop
    xc, yc, R, residu = calibrate.leastsq_circle(np.real(fine_z), np.imag(fine_z))
    # compute angle between (off_res,off_res),(0,0) and (off_ress,off_res),(xc,yc) of the the fitted circle
    off_res_i, off_res_q = (np.real(fine_z[0]) + np.real(fine_z[-1])) / 2., (
            np.imag(fine_z[0]) + np.imag(fine_z[-1])) / 2.
    x1, y1, = -off_res_i, -off_res_q
    x2, y2 = xc - off_res_i, yc - off_res_q
    dot = x1 * x2 + y1 * y2  # dot product
    det = x1 * y2 - y1 * x2  # determinant
    angle = np.arctan2(det, dot)
    phi_guess = angle

    # if phi is large better re guess f0
    # f0 should be the farthers from the off res point
    if (np.abs(phi_guess) > 0.3):
        dist1 = np.sqrt((np.real(fine_z[0]) - np.real(fine_z)) ** 2 + (np.imag(fine_z[0]) - np.imag(fine_z)) ** 2)
        dist2 = np.sqrt((np.real(fine_z[-1]) - np.real(fine_z)) ** 2 + (np.imag(fine_z[-1]) - np.imag(fine_z)) ** 2)
        fr_guess_index = np.argmax((dist1 + dist2))
        fr_guess = fine_x[fr_guess_index]
        fine_z_derot = (fine_z - (off_res_i + 1.j * off_res_q)) * np.exp(1j * (-phi_guess)) + (
                off_res_i + 1.j * off_res_q)
        # fr_guess_index = np.argmin(np.abs(fine_z_derot))
        # fr_guess = fine_x[fr_guess_index]
        mag_max = np.max(np.abs(fine_z_derot) ** 2)
        mag_min = np.min(np.abs(fine_z_derot) ** 2)
        mag_3dB = (mag_max + mag_min) / 2.
        half_distance = np.abs(fine_z_derot) ** 2 - mag_3dB
        right = half_distance[np.argmin(np.abs(fine_z_derot)):-1]
        left = half_distance[0:np.argmin(np.abs(fine_z_derot))]
        right_index = np.argmin(np.abs(right)) + np.argmin(np.abs(fine_z_derot))
        left_index = np.argmin(np.abs(left))
        Q_guess_Hz = fine_x[right_index] - fine_x[left_index]
        Q_guess = fr_guess / Q_guess_Hz
        # also fix amp guess
        d = np.max(20 * np.log10(np.abs(gain_z))) - np.min(20 * np.log10(np.abs(fine_z_derot)))
        amp_guess = 0.0037848547850284574 + 0.11096782437821565 * d - 0.0055208783469291173 * d ** 2 + 0.00013900471000261687 * d ** 3 + -1.3994861426891861e-06 * d ** 4

    # guess non-linearity parameter
    # might be able to guess this by taking the ratio of the distance between min and max distance between
    # iq points in fine sweep
    a_guess = 0

    # b0 and b1 guess
    xlin = (gain_x - fr_guess) / fr_guess
    b1_guess = (np.abs(gain_z)[-1] ** 2 - np.abs(gain_z)[0] ** 2) / (xlin[-1] - xlin[0])
    b0_guess = np.max((np.max(np.abs(fine_z) ** 2), np.max(np.abs(gain_z) ** 2)))

    # cable delay guess tau
    # y = mx +b
    # m = (y2 - y1)/(x2-x1)
    # b = y-mx
    m = (gain_phase - np.roll(gain_phase, 1)) / (gain_x - np.roll(gain_x, 1))
    b = gain_phase - m * gain_x
    m_best = np.median(m[~np.isnan(m)])
    tau_guess = m_best / (2 * np.pi)

    if verbose:
        print("fr guess  = %.3f MHz" % (fr_guess / 10 ** 6))
        print("Q guess   = %.2f kHz, %.1f" % ((Q_guess_Hz / 10 ** 3), Q_guess))
        print("amp guess = %.2f" % amp_guess)
        print("phi guess = %.2f" % phi_guess)
        print("b0 guess  = %.2f" % b0_guess)
        print("b1 guess  = %.2f" % b1_guess)
        print("tau guess = %.2f x 10^-7" % (tau_guess / 10 ** -7))

    x0 = [fr_guess, Q_guess, amp_guess, phi_guess, a_guess, b0_guess, b1_guess, fr_guess]
    return x0


def guess_so_resonator_cable(freqs, real, imag):
    """
    Fitting function that takes in frequency and real and imaginary parts of the
    transmission of a resonator (needs to trimmed down to only data for a
    single resonator) and returns fitted parameters to the ``resonator_cable``
    model.
    Args
    ----
    freqs : float ndarray
        Frequencies that line up with complex transmission data.
    real : float ndarray
        Real part of resonator complex transmission to be fit
    imag : float ndarray
        Imaginary part of resonator complex transmission to be fit.
    Returns
    -------
    result : tuple
        Initial guess parameters for a fitting of so_resonator_cable model, the 9 values expected are:
        (f_0_guess, Q_guess, Q_e_real_guess, Q_e_imag_guess, delay, phi_offset, fmin, A_mag, A_mag_slope)
    """
    s21_complex = np.vectorize(complex)(real, imag)

    # set our initial guesses
    argmin_s21 = np.abs(s21_complex).argmin()
    fmin = freqs.min()
    fmax = freqs.max()
    f_span = fmax - fmin
    f_0_guess = freqs[argmin_s21]

    Q_min = 0.1 * (f_0_guess / (fmax - fmin))
    delta_f = np.diff(freqs)
    min_delta_f = delta_f[delta_f > 0].min()
    Q_max = f_0_guess / min_delta_f
    Q_guess = np.sqrt(Q_min * Q_max)

    s21_min = np.abs(s21_complex[argmin_s21])
    s21_max = np.abs(s21_complex).max()
    Q_e_real_guess = Q_guess / (1 - s21_min / s21_max)
    Q_e_imag_guess = 0.0

    A_slope, A_offset = np.polyfit(freqs - fmin, np.abs(s21_complex), 1)
    A_mag_guess = A_offset
    A_slope_guess = A_slope / A_mag_guess
    phi_slope, phi_guess = np.polyfit(freqs - fmin, np.unwrap(np.angle(s21_complex)), 1)
    delay_guess = -phi_slope / (2 * np.pi)
    f_min_guess = fmin


    n_inf = float('-inf')
    inf = float('inf')
    phi_min = phi_guess - np.pi
    phi_max = phi_guess + np.pi
    fmin_min = fmin - f_span

    x0 = (f_0_guess, Q_guess, Q_e_real_guess, Q_e_imag_guess, delay_guess, phi_guess, f_min_guess, A_mag_guess, A_slope_guess)

    bounds = ((fmin,   Q_min,          1.0e0,         -1.0e7,      n_inf,    phi_min,  fmin_min,        n_inf,             n_inf),
              (fmax,   Q_max,          1.0e7,          1.0e7,        inf,    phi_max,      fmax,          inf,               inf))

    return x0, bounds


def calc_qc_qi(qr: float, amp: float):
    qc = qr / amp
    qi = 1.0 / ((1.0 / qr) - (1.0 / qc))
    return qc, qi


default_null_strings = {'', 'None', 'none', 'null', 'NaN', 'nan'}
default_true_strings = {'Y', 'y', 'True', 'true'}
default_false_strings = {'N', 'n', 'False', 'false'}


def autoscale_from_data(data, log_scale=False, pad_fraction=0.05, min_data_span=1.0e-8):
    data_min = min(data)
    data_max = max(data)
    if log_scale:
        data_min = np.log10(data_min)
        data_max = np.log10(data_max)
    data_span = data_max - data_min
    if data_span == 0.0:
        data_span = min_data_span
    pad = data_span * pad_fraction
    plot_min = data_min - pad
    plot_max = data_max + pad
    if log_scale:
        plot_min = 10 ** plot_min
        plot_max = 10 ** plot_max
    return plot_min, plot_max


def make_int(test_num_str):
    """
    Parameters
    ----------
    test_num_str:
        str, required. A string to test to see it can be cast into and integer.
    Returns int, if the string can be cast into and integer, otherwise it returns the original string.
    -------
    """
    try:
        return int(test_num_str)
    except ValueError:
        return test_num_str


def make_num(test_datum):

    # tests to se if this string is an int
    test_datum_maybe_int = make_int(test_num_str=test_datum)
    if isinstance(test_datum_maybe_int, int):
        return test_datum_maybe_int
    else:
        # either a float or a sting.
        try:
            return float(test_datum_maybe_int)
        except ValueError:
            return test_datum_maybe_int


def format_datum(test_datum, null_strs=None, true_strs=None, false_strs=None):
    # strip off spaces and newline charters
    test_datum_stripped = test_datum.strip()
    # Null handling
    if null_strs is None:
        null_strs = default_null_strings
    if not isinstance(null_strs, set):
        null_strs = set(null_strs)
    # True handling
    if true_strs is None:
        true_strs = default_true_strings
    if not isinstance(true_strs, set):
        true_strs = set(true_strs)
    # False handling
    if false_strs is None:
        false_strs = default_false_strings
    if not isinstance(false_strs, set):
        false_strs = set(false_strs)
    # see if this an expected null, true, or false value, else move on to number testing
    if test_datum_stripped in null_strs:
        return None
    elif test_datum_stripped in true_strs:
        return True
    elif test_datum_stripped in false_strs:
        return False
    else:
        return make_num(test_datum=test_datum_stripped)


def line_format(raw_line, delimiter=','):
    return [format_datum(datum_raw) for datum_raw in raw_line.split(delimiter)]


# https://stackabuse.com/how-to-print-colored-text-in-python/
style_to_number = {'normal': 0, 'bold': 1, 'dark': 2, 'light': 3, 'underline': 4, 'blink': 5}
text_color_to_number = {'black': 30, 'red': 31, 'green': 32, 'yellow': 33, 'blue': 34, 'purple': 35,
                        'cyan': 36, 'white': 37}
background_color_to_number = {'black': 40, 'red': 41, 'green': 42, 'yellow': 43, 'blue': 44, 'purple': 45,
                              'cyan': 46, 'white': 47}
text_color_matplotlib = {'black': 'white', 'red': 'black', 'green': 'white', 'yellow': 'black', 'blue': 'white',
                         'purple': 'white', 'cyan': 'black', 'white': 'black'}


def colorize_text(text: str, style_text: str = 'normal', color_text: str = 'white', color_background: str = 'black'):
    if style_text not in style_to_number.keys():
        raise ValueError(f"Style '{style_text}' not found, options are {style_to_number.keys()}")
    if color_text not in text_color_to_number.keys():
        raise ValueError(f"Color '{color_text}' not found, options are {text_color_to_number.keys()}")
    if color_background not in background_color_to_number.keys():
        raise ValueError(f"Color '{color_background}' not found, options are {background_color_to_number.keys()}")
    style_number = style_to_number[style_text]
    text_color_number = text_color_to_number[color_text]
    background_color_number = background_color_to_number[color_background]
    return f"\033[{style_number};{text_color_number};{background_color_number}m{text}\033[0;0m"


def derived_text(text):
    return colorize_text(text, style_text='bold', color_text='black', color_background='green')


def filename_text(text):
    return colorize_text(text, style_text='bold', color_text='white', color_background='black')


def write_text(text):
    return colorize_text(text, style_text='bold', color_text='black', color_background='purple')



from submm.multitone_kidPy import analyze

skip_beginning = 10 #skip first few streaming points

# read in fine and gain scans
fine_filename = "1521699164-Mar-22-2018-00-12-44.dir"
gain_filename = "1521699121-Mar-22-2018-00-12-01.dir"
stream_filename = "1521699068-Mar-22-2018-00-11-08.dir"

cal_dict = analyze.calibrate_multi(fine_filename, gain_filename, stream_filename, skip_beginning = skip_beginning)

psd_dict = analyze.noise_multi(cal_dict)



'''
Spectrum analyzer object for remote reading out of a
Anritsu MS2711D handheld spectrum analyzer
Author Jordan Wheeler Wheeler1711@gmail.com

Note: commands must be entered in remote mode and then to
resume sweeping you must exit remote mode
enter_remote_mode waits until the sweep is finished and thus 
the serial timeout must be longer than the sweep time 
                               Span    
sweep time is propotional to ----------   
                              RBW*VBW          
'''
#check for usb to serial with
# ls tty.usbserial*
# written for python3
# requires pyserial and numpy
# pip install pyserial

import serial
import numpy as np
import struct

class spectrum_analyzer(object):

    def __init__(self):
        self.ser = serial.Serial('/dev/tty.usbserial') # mac usb to serial converter Windows COM3 or something
        #print(self.ser.name)
        self.ser.timeout = 30

    def check_response_for_error(self,rx): # standard response from ms2711d to confirm data tranmission
        if struct.unpack('B',rx)[0] == 255: #success
            return True
        elif struct.unpack('B',rx)[0] == 224:
            print("invaled parameter")
            return False
        elif struct.unpack('B',rx)[0] == 238:
            print("time out error")
            return False
        else:
            print("unknown error")
            return False

    def enter_remote_mode(self):
        SerialCommand = struct.pack('B', 69) # 69 is the Enter Remote Mode serial command
        self.ser.write(SerialCommand)
        rx = self.ser.read(13)
        if len(rx) == 13:
            return True
        else:
            return False

    def exit_remote_mode(self):
        SerialCommand = struct.pack('B', 255) # 255 is the exit Remote Mode serial command
        self.ser.write(SerialCommand)
        rx = self.ser.read(1)
        #print(rx)
        if len(rx) == 1:
            return True
        else:
            return False

    def grab_last_sweep_command(self):
        # there is a bunch of stuff in the returned byte array
        # see programming manual for details
        # should proabably grab more info return a dictionary
        SerialCommand = struct.pack('B', 33) # grab sweep
        SerialCommand2 = struct.pack('B', 0) #last sweep
        self.ser.write(SerialCommand)
        self.ser.write(SerialCommand2)
        rx = self.ser.read(2035)

        # get frequencies
        start_freq_bytes = rx[56:60]
        stop_freq_bytes =rx[60:64]
        freq_scale_factor_bytes = rx[334:336]
        freq_scale_factor = struct.unpack('>H',freq_scale_factor_bytes )[0]
        start_freq = struct.unpack('>I',start_freq_bytes )[0]/freq_scale_factor
        stop_freq = struct.unpack('>I',stop_freq_bytes )[0]/freq_scale_factor
        freqs = np.linspace(start_freq,stop_freq,401)

        # get data
        if len(rx) != 2035:
            return False
        else:
            #data is 432 to 2035 4 bytes per point
            data = np.zeros(401)
            for i in range(0,401):
                data[i] = (struct.unpack('>i', rx[431+i*4:435+i*4])[0]-270000)/1000

            return freqs, data

    def grab_last_sweep(self): # wrapper includes enter exit remote mode
        self.enter_remote_mode()
        freqs,data = self.grab_last_sweep_command()
        self.exit_remote_mode()
        return freqs, data

    def set_resolution_bandwidth_command(self,bandwidth):
        bandwidth_options = np.asarray((100,300,1000,3000,10000,30000,100000,300000,1000000)) #only some bandwiths are availible
        if ~np.any(bandwidth == bandwidth_options):
            print("Please select a proper bandwidth option i.e.")
            print(bandwidth_options)
            return
        else:
            SerialCommand = struct.pack('B', 141) # set bandwidth
            SerialCommand2 = struct.pack('>I', bandwidth) #unsiged int
            self.ser.write(SerialCommand)
            self.ser.write(SerialCommand2)
            rx = self.ser.read(1)
            return self.check_response_for_error(rx)  # success?   

    def set_resolution_bandwidth(self,bandwidth): # wrapper includes enter exit remote mode
        self.enter_remote_mode()
        success = self.set_resolution_bandwidth_command(bandwidth)
        self.exit_remote_mode()
        return success

    def set_start_and_stop_frequency_command(self,start,stop):
        # must be integer
        SerialCommand = struct.pack('B', 99) # set start and stop frequency
        # in Hertz
        SerialCommand2 = struct.pack('>I', start) #unsigned int
        SerialCommand3 = struct.pack('>I', stop)
        self.ser.write(SerialCommand)
        self.ser.write(SerialCommand2)
        self.ser.write(SerialCommand3)
        rx = self.ser.read(1)
        return self.check_response_for_error(rx)

    def set_start_and_stop_frequency(self,start,stop): # wrapper includes enter exit remote mode
        self.enter_remote_mode()
        success = self.set_start_and_stop_frequency_command(start,stop)
        self.exit_remote_mode()
        return success

    def set_attenuation_command(self,atten):
        if atten < 0:
            print("attenuation value too low")
            return False
        elif atten>51:
            print("attenuation value too high")
            return False
        else:
            pass #just right
        
        SerialCommand = struct.pack('B', 143) # set start and stop frequency
        SerialCommand2 = struct.pack('B', atten) #unsigned char
        self.ser.write(SerialCommand)
        self.ser.write(SerialCommand2)
        rx = self.ser.read(1)
        return self.check_response_for_error(rx)

    def set_attenuation(self,atten):
        self.enter_remote_mode()
        success = self.set_attenuation_command(atten)
        self.exit_remote_mode()
        return success



import numpy as np
from scipy import optimize
import matplotlib.pyplot as plt
from scipy import fftpack
from scipy import interpolate

# this script is for calibrating a kinetic inductance detector to convert
# changes in i and q to the shift of the resonator

def calc_R(x,y, xc, yc):
    """ calculate the distance of each 2D points from the center (xc, yc) """
    return np.sqrt((x-xc)**2 + (y-yc)**2)

def f(c, x, y):
    """ calculate the algebraic distance between the data points and the mean circle centered at c=(xc, yc) """
    Ri = calc_R(x, y, *c)
    return Ri - Ri.mean()

def leastsq_circle(x,y):
    # coordinates of the barycenter
    x_m = np.mean(x)
    y_m = np.mean(y)
    center_estimate = x_m, y_m
    center, ier = optimize.leastsq(f, center_estimate, args=(x,y))
    xc, yc = center
    Ri       = calc_R(x, y, *center)
    R        = Ri.mean()
    residu   = np.sum((Ri - R)**2)
    return xc, yc, R, residu

def plot_data_circle(x,y, xc, yc, R):
    theta_fit = np.linspace(-np.pi, np.pi, 180)
    x_fit = xc + R*np.cos(theta_fit)
    y_fit = yc + R*np.sin(theta_fit)
    plt.plot(x_fit, y_fit, 'b-' , label="fitted circle", lw=1)

def fit_cable_delay(gain_f,gain_phase):
    #fitting is complicated when phase wraps from
    # positive pi to negative pi
    # so we apply a phase offset ot get the data to be centered at 0 phase
    # first we move from -pi to pi to 0 to 2pi so we can use a modulus function
    # then we shift the phase so that the center point in on pi
    # gain f in Hz
    gain_phase = np.mod(gain_phase+np.pi-((gain_phase[len(gain_phase)//2]+np.pi)-np.pi),2*np.pi)
    #shift back to -pi to pi space for fun
    gain_phase = gain_phase -np.pi
    p_phase = np.polyfit(gain_f,gain_phase,1)
    tau = p_phase[0]/(2.*np.pi)
    poly_func_phase = np.poly1d(p_phase)
    fit_data_phase = poly_func_phase(gain_f)

    return tau,fit_data_phase,gain_phase

def fit_cable_delay_from_slope(f,phase,plot = True):
    # gain f in Hz
    phase_gradient = np.gradient(phase)/np.gradient(f)/2/np.pi
    tau = np.median(phase_gradient)
    if plot:
        plt.plot(f/10**6,phase_gradient,label = "data")
        plt.plot(f/10**6,tau*np.ones(len(f)),label = "fit")
        plt.xlabel("Frequency (MHz)")
        plt.ylabel("Phase gradient")
        plt.legend()
        plt.title("Tau = "+str(tau))
        plt.show()
    return tau,phase_gradient
        
def remove_cable_delay(f,z,tau):
    z_corr = z*np.exp(2j*np.pi*tau*f)
    return z_corr

def fft_noise(z_stream,df_over_f,sample_rate):
    npts_fft = int(2**(np.floor(np.log2(df_over_f.size)))) 
    Sxx = 2*fftpack.fft(df_over_f,n = npts_fft)*np.conj(fftpack.fft(df_over_f,n = npts_fft))/sample_rate*npts_fft/npts_fft**2
    S_per = 2*fftpack.fft(np.real(z_stream),n = npts_fft)*np.conj(fftpack.fft(np.real(z_stream),n = npts_fft))/sample_rate*npts_fft/npts_fft**2
    S_par = 2*fftpack.fft(np.imag(z_stream),n = npts_fft)*np.conj(fftpack.fft(np.imag(z_stream),n = npts_fft))/sample_rate*npts_fft/npts_fft**2
    fft_freqs = fftpack.fftfreq(npts_fft,1./sample_rate)
    return fft_freqs,Sxx,S_per,S_par

def interp_phase_to_df_over_f(phase_fine,phase_stream,iq_sweep_freqs,extrap = "min_max"):
    #interp phase to frequency
    if extrap == "min_max":
        fill_value = (iq_sweep_freqs[-1],iq_sweep_freqs[0])
    else:
        fill_value = "extrapolate"
        
    f_interp = interpolate.interp1d(phase_fine, iq_sweep_freqs,kind = 'quadratic',
                                    bounds_error = False,fill_value = fill_value)
    freqs_stream = f_interp(phase_stream)
    mean_freq = np.mean(freqs_stream)
    stream_df_over_f = freqs_stream/mean_freq-1.
    return stream_df_over_f,f_interp



#!/usr/bin/env python3

from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
from numpy import fft
import matplotlib.pyplot as plt
from scipy.stats import binned_statistic
from scipy import signal


def PCA_SVD(orig_array, n_comp_remove, plot=False,sample_rate =488.28125,
            outfile_dir = "./",plot_decimate = 1 ):
    """This is a Principal Component Analysis cleaning of common modes from the 
    data. This is the implementation using Single Value Decomposition which is
    usually held as a better implementation, but may be undesireable for long
    time streams

    array is the 2D numpy array with the ith time point, and jth detector in
        array[i,j]
    n_comp_remove is the number of components to remove as the common mode
    """

    #subtract off the mean of each time stream
    mean_array = np.mean(orig_array, axis=0)
    array = orig_array - mean_array

    #normalize by variance (ensure the channels are weighted evenly rather
    #than any biasing towards the higher response channels
    var_array = np.var(array, axis=0)

    hold_res = np.copy(array)
    array = array[:, var_array !=0.] #bad calibration can create timestreams of
            #all 1 or 0, meaning variance normailization does not work
            #just ignore these channels and keep them unchanged
    array /= np.sqrt(var_array[var_array != 0])

    #convert to the transverse array since usual notation and equations have
    #array[j,i] where i is the trial/timepoint and j is the detector
    array = array.T
    #compute the svd, not using full matrices (drops zero columns which only
    #use memory and prevent easy matrix multiplication)
    U, S, Vh = np.linalg.svd(array, full_matrices=False, compute_uv=True)
    removed = Vh[0:n_comp_remove,:] #save the removed components so can be
        #inspected later, svd should sort components in decending strength
    S[0:n_comp_remove] = 0. #set the components being removed to zero
    #just remultiply the arrays together to get the results with the removed
    #components explicitely set to zero
    cleaned_array = np.dot(U * S, Vh)

    #undo the transvers and variance normalization and add back the mean in
    #case we are working in I, Q or some other case where a DC offset is
    #necessary
    cleaned_array = cleaned_array.T
    #to keep ignored channels the same, just put cleaned data into copy of 
    #original array
    hold_res[:,var_array != 0.] = cleaned_array
    cleaned_array = hold_res
    #undo normalizations and mean subtraction
    cleaned_array *= np.sqrt(var_array)
    cleaned_array += mean_array
    delta = orig_array - cleaned_array
    if plot:
        plot_PCA(U, S, Vh, var_array, mean_array,sample_rate = sample_rate,
                 outfile_dir = outfile_dir,decimate = plot_decimate)

    #return both the cleaned array and the components removed
    
    return cleaned_array, removed.T


def PCA_covariance(array, n_comp_remove):
    """This is a Principal Component Analysis cleaning of common modes from the
    data. This is using the eigenvectors of the covariance matrix to find the
    principal components.

    array is a 2D numpy array with the ith time point and jth detector in
        array[i,j]
    n_comp_remove is the number of components to remove as the common mode
    """

    #subtract off the mean_of each time stream
    mean_array = np.mean(array, axis=0)
    array -= mean_array

    #normalize each channel by its variance
    var_array = np.var(array, axis=0)
    array /= var_array

    #calculate the covariance matrix
    cov_array = np.dot(array.T, array)

    #calculate eigen vals/vectors then sort in descending order
    eigen_val, eigen_vec = np.linalg.eig(cov_array)
    
    sort_index = np.argsort(eigen_val)[::-1]
    eigen_vec = eigen_vec[:, sort_index]
    eigen_val = eigen_val[sort_index]

    #use eigen vectors to project into PCA components
    component_streams = np.dot(array, eigen_vec)


def plot_PCA(U, S, Vh, variance, mean,sample_rate = 488.28125,outfile_dir = "./",decimate = 1):
    """For diagnostics and understanding what the PCA is removing, this 
    plots the timestream and psd of every principal component, along with the 
    mixing matrix (how the components map onto the original timestreams
    """
    mapping_matrix = U * S
    pdf_pages = PdfPages(outfile_dir+'PCA_plots.pdf')

    fig = plt.figure()
    plt.suptitle('Mixing Matrix')
    mat_plot = plt.imshow(np.real(mapping_matrix))
    fig.colorbar(mat_plot)
    pdf_pages.savefig(fig)
    plt.close(fig)

    
    t_vals = np.linspace(0, Vh.shape[1] / sample_rate, Vh.shape[1])
    fft_freqs = fft.fftfreq(Vh.shape[1],1./sample_rate)
    for i in range(Vh.shape[0]):
        component_psd = 2. * fft.fft(Vh[i,:]) * np.conj(fft.fft(Vh[i,:])
                    ) / (sample_rate * Vh.shape[1])
        fig = plt.figure(i + 1000,figsize = (16,6))
        plt.subplot(211)
        plt.title("Timestream of Principal component {0}".format(i))
        factors_of_10 = int(np.floor(np.log10(decimate)))
        Vh_decimated = Vh[i,:]
        for k in range(0,factors_of_10): # not suppose to decimate all at once
            Vh_decimated = signal.decimate(Vh_decimated,10,axis = 0)
            
        Vh_decimated = signal.decimate(Vh_decimated,decimate//(10**factors_of_10),axis =0)
        #Vh_decimated = signal.decimate(Vh[i,:],decimate)
        t_vals_decimated = np.arange(0,Vh_decimated.shape[0])*1/sample_rate*decimate
        plt.plot(t_vals_decimated ,np.real(Vh_decimated),linewidth = 2)
        plt.ylabel("response (A. U.)")
        plt.xlabel("Time (s)")
        plt.subplot(212)
        plot_bins = np.logspace(np.log10(np.min(np.abs(fft_freqs[1:]))),np.log10(np.max(fft_freqs)),1000)
        binned_freq =  binned_statistic(fft_freqs, fft_freqs, bins=plot_bins)[0] #bin the frequecy against itself 
        binned_psd = binned_statistic(fft_freqs, np.abs(component_psd), bins=plot_bins)[0]
        nan_index = np.isnan(binned_freq)
        plt.loglog(binned_freq[~nan_index] ,binned_psd[~nan_index], linewidth = 2)
        WNL = np.mean(np.abs(component_psd[fft_freqs > 20.]))
        plt.ylim([0.005 * WNL, max(binned_psd[~nan_index])])
        plt.ylabel("response (A. U. / Hz)")
        plt.xlabel("Frequency (Hz)")

        pdf_pages.savefig(fig)
        plt.close(fig)
    pdf_pages.close()




import numpy as np
import matplotlib.pyplot as plt
from submm.KIDs import calibrate
from submm.KIDs.res.utils import amplitude_normalization_sep
from scipy import interpolate
import pickle
from scipy.stats import binned_statistic


def calibrate_single_tone(fine_f, fine_z, gain_f, gain_z, stream_f, stream_z, plot_period=1, interp="quadratic"):
    fig = plt.figure(3, figsize=(16, 10))

    plt.subplot(241, aspect='equal')
    plt.title("Raw data")
    plt.plot(np.real(stream_z[::plot_period]), np.real(stream_z[::plot_period]), '.')
    plt.plot(np.real(fine_z), np.imag(fine_z), 'o')
    plt.plot(np.real(gain_z), np.imag(gain_z), 'o')

    plt.subplot(242)
    plt.title("Raw data")
    plt.plot(np.real(stream_z[::plot_period]), np.imag(stream_z[::plot_period]), '.')
    plt.plot(np.real(fine_z), np.imag(fine_z), 'o')

    # normalize amplitude variation in the gain scan
    amp_norm_dict = amplitude_normalization_sep(gain_f, gain_z, fine_f, fine_z, stream_f, stream_z)

    plt.subplot(243)
    plt.title("Gain amplitude variation fit")
    plt.plot(gain_f, 10 * np.log10(np.abs(gain_z) ** 2), 'o')
    plt.plot(gain_f, 10 * np.log10(np.abs(amp_norm_dict['normalized_gain']) ** 2), 'o')
    plt.plot(fine_f, 10 * np.log10(np.abs(amp_norm_dict['normalized_fine']) ** 2), 'o')
    plt.plot(gain_f, 10 * np.log10(np.abs(amp_norm_dict['poly_data']) ** 2))

    plt.subplot(244)
    plt.title("Data normalized for gain amplitude variation")
    plt.plot(np.real(amp_norm_dict['normalized_fine']), np.imag(amp_norm_dict['normalized_fine']), 'o')
    # plt.plot(gain_dict['freqs'][:,k]*10**6,np.log10(np.abs(amp_norm_dict['poly_data'])**2))
    plt.plot(np.real(amp_norm_dict['normalized_stream'][::plot_period]),
             np.imag(amp_norm_dict['normalized_stream'][::plot_period]), '.')
    # fit the gain
    gain_phase = np.arctan2(np.real(amp_norm_dict['normalized_gain']), np.imag(amp_norm_dict['normalized_gain']))
    tau, fit_data_phase, gain_phase_rot = calibrate.fit_cable_delay(gain_f, gain_phase)

    plt.subplot(245)
    plt.title("Gain phase fit")
    plt.plot(gain_f, gain_phase_rot, 'o')
    plt.plot(gain_f, fit_data_phase)
    plt.xlabel("Frequency (MHz)")
    plt.ylabel("Phase")

    # remove cable delay
    gain_corr = calibrate.remove_cable_delay(gain_f, amp_norm_dict['normalized_gain'], tau)
    fine_corr = calibrate.remove_cable_delay(fine_f, amp_norm_dict['normalized_fine'], tau)
    stream_corr = calibrate.remove_cable_delay(stream_f, amp_norm_dict['normalized_stream'], tau)

    plt.subplot(246)
    plt.title("Cable delay removed")
    plt.plot(np.real(gain_corr), np.imag(gain_corr), 'o')
    plt.plot(np.real(fine_corr), np.imag(fine_corr), 'o')
    plt.plot(np.real(stream_corr)[10:-10][::plot_period], np.imag(stream_corr)[10:-10][::plot_period], '.')

    # fit a cicle to the data
    xc, yc, R, residu = calibrate.leastsq_circle(np.real(fine_corr), np.imag(fine_corr))

    # move the data to the origin
    gain_corr = gain_corr - xc - 1j * yc
    fine_corr = fine_corr - xc - 1j * yc
    stream_corr = stream_corr - xc - 1j * yc

    # rotate so streaming data is at 0 pi
    phase_stream = np.arctan2(np.imag(stream_corr), np.real(stream_corr))
    med_phase = np.median(phase_stream)

    gain_corr = gain_corr * np.exp(-1j * med_phase)
    fine_corr = fine_corr * np.exp(-1j * med_phase)
    stream_corr = stream_corr * np.exp(-1j * med_phase)

    plt.subplot(247)
    plt.title("Moved to 0,0 and rotated")
    plt.plot(np.real(stream_corr)[2:-1][::plot_period], np.imag(stream_corr)[2:-1][::plot_period], '.')
    plt.plot(np.real(gain_corr), np.imag(gain_corr), 'o')
    plt.plot(np.real(fine_corr), np.imag(fine_corr), 'o')
    calibrate.plot_data_circle(np.real(fine_corr) - xc, np.imag(fine_corr) - yc, 0, 0, R)

    phase_fine = np.arctan2(np.imag(fine_corr), np.real(fine_corr))
    use_index = np.where((-np.pi / 2. < phase_fine) & (phase_fine < np.pi / 2))
    phase_stream = np.arctan2(np.imag(stream_corr), np.real(stream_corr))

    # interp phase to frequency
    f_interp = interpolate.interp1d(phase_fine, fine_f, kind=interp, bounds_error=False, fill_value=0)

    phase_small = np.linspace(np.min(phase_fine), np.max(phase_fine), 1000)
    freqs_stream = f_interp(phase_stream)
    stream_df_over_f_all = stream_df_over_f = freqs_stream / np.mean(freqs_stream) - 1.

    plt.subplot(248)
    plt.plot(phase_fine, fine_f, 'o')
    plt.plot(phase_small, f_interp(phase_small), '--')
    plt.plot(phase_stream[::plot_period], freqs_stream[::plot_period], '.')
    plt.ylim(np.min(freqs_stream) - (np.max(freqs_stream) - np.min(freqs_stream)) * 3,
             np.max(freqs_stream) + (np.max(freqs_stream) - np.min(freqs_stream)) * 3)
    plt.xlim(np.min(phase_stream) - np.pi / 4, np.max(phase_stream) + np.pi / 4)
    plt.xlabel("phase")
    plt.ylabel("Frequency")

    plt.savefig("calibration.pdf")

    cal_dict = {'fine_z': fine_z,
                'gain_z': gain_z,
                'stream_z': stream_z,
                'fine_freqs': fine_f,
                'gain_freqs': gain_f,
                'stream_corr': stream_corr,
                'gain_corr': gain_corr,
                'fine_corr': fine_corr,
                'stream_df_over_f': stream_df_over_f_all}

    pickle.dump(cal_dict, open("cal.p", "wb"), 2)
    return cal_dict


def noise(cal_dict, sample_rate, title=None):
    fft_freqs, Sxx, S_per, S_par = calibrate.fft_noise(cal_dict['stream_corr'], cal_dict['stream_df_over_f'],
                                                       sample_rate)
    plot_bins = np.logspace(-3, np.log10(250000), 1000)
    binnedfreq = binned_statistic(fft_freqs, fft_freqs, bins=plot_bins)[0]  # bin the frequecy against itself
    binnedpsd = binned_statistic(fft_freqs, np.abs(Sxx), bins=plot_bins)[0]

    binnedper = binned_statistic(fft_freqs, np.abs(S_per), bins=plot_bins)[0]
    binnedpar = binned_statistic(fft_freqs, np.abs(S_par), bins=plot_bins)[0]
    amp_subtracted = np.abs(binnedpsd) * (binnedpar - binnedper) / binnedpar

    fig = plt.figure(4, figsize=(16, 6))
    plt.subplot(122)
    if title is None:
        plt.title("Sxx")
    else:
        plt.title('Sxx ' + title)
    # plt.loglog(fft_freqs,np.abs(Sxx))
    plt.loglog(binnedfreq, np.abs(binnedpsd), linewidth=2, label="Sxx raw")
    plt.loglog(binnedfreq, amp_subtracted, linewidth=2, label="raw amp subtracted")
    # plt.ylim(10**-18,10**-15)
    plt.ylabel("Sxx (1/Hz)")
    plt.xlabel("Frequency (Hz)")
    plt.legend()

    plt.subplot(121)
    # plt.loglog(fft_freqs,S_per)
    # plt.loglog(fft_freqs,S_par)
    plt.loglog(binnedfreq, binnedper, label="amp noise")
    plt.loglog(binnedfreq, binnedpar, label="detect noise")
    plt.legend()
    # plt.ylim(10**2,10**6)
    plt.xlabel("Frequency (Hz)")

    plt.savefig("psd.pdf")

    psd_dict = {'fft_freqs': fft_freqs,
                'Sxx': Sxx,
                'S_per': S_per,
                'S_par': S_par,
                'binned_freqs': binnedfreq,
                'Sxx_binned': binnedpsd,
                'S_per_binned': binnedper,
                'S_par_binned': binnedpar,
                'amp_subtracted': amp_subtracted}

    # save the psd dictionary
    pickle.dump(psd_dict, open("psd.p", "wb"), 2)

    return psd_dict



import numpy as np
import matplotlib.pyplot as plt
from submm.multitone_mako import read_multitone
from submm.KIDs.res.fitting import fit_nonlinear_iq_with_err

# change log
# JDW 2017-08-17 changed to include amplitude normalization option

res_num = 0

cal = read_multitone.readcal("mako20161007_223744Cal.fits")
gain = read_multitone.readcal("mako20161007_224502Cal.fits")

all_i = np.vstack((cal['I'], gain['I']))
all_q = -np.vstack(
    (cal['Q'], gain['Q']))  # notice the negative for the Q the mako multitone seems to have the wrong sign on Q
all_freqs = np.vstack((cal['freqs'], gain['freqs']))

x = all_freqs[:, res_num]
z = all_i[:, res_num] + 1.0j * all_q[:, res_num]

# multitone glitches
x[30] = np.nan
x[34] = np.nan
x[50] = np.nan
x[34] = np.nan
z[30] = np.nan
z[34] = np.nan
z[50] = np.nan
z[34] = np.nan

# fitter does not like nans
z_nonan = z[~np.isnan(z)]
x_nonan = x[~np.isnan(z)]

f0_guess = x_nonan[np.argmin(np.abs(z_nonan))]

# try with or without these
x0 = [f0_guess, 10000., 0.37, np.pi * -0.1, .6, np.real(z_nonan[60]), np.imag(z_nonan[60]), 3 * 10 ** -7, f0_guess]
bounds = ([np.min(x_nonan), 2000, .01, -4.0 * np.pi, 0, -5, -5, 1 * 10 ** -9, np.min(x_nonan)],
          [np.max(x_nonan), 200000, 100, 4.0 * np.pi, 5, 5, 5, 1 * 10 ** -6, np.max(x_nonan)])

fit_dict = fit_nonlinear_iq_with_err(x_nonan, z_nonan, x0=x0, bounds=bounds, amp_norm=True)  # with defined bounds
# fit_dict = resonance_fitting.fit_nonlinear_iq_with_err( x_nonan, z_nonan) #with out defined bounds

fit = fit_dict['fit']
fit_result = fit_dict['fit_result']
x0_result = fit_dict['x0_result']

plt.figure(1)
plt.plot(np.real(x0_result), np.imag(x0_result), '+', label="x0")
plt.plot(np.real(z_nonan), np.imag(z_nonan), 'o', label="data")
plt.plot(np.real(fit_dict['z']), np.imag(fit_dict['z']), 'o', label="data after amp_norm")
plt.plot(np.real(fit_result), np.imag(fit_result), '*', label="fit")
plt.legend(loc=4)

plt.figure(2)
plt.plot(x_nonan, np.sqrt(np.real(z_nonan) ** 2 + np.imag(z_nonan) ** 2), 'o', label="data")
plt.plot(x_nonan, np.sqrt(np.real(fit_dict['z']) ** 2 + np.imag(fit_dict['z']) ** 2), 'o', label="data after amp norm")
plt.plot(x_nonan, np.sqrt(np.real(x0_result) ** 2 + np.imag(x0_result) ** 2), '+', label="x0")
plt.plot(x_nonan, np.sqrt(np.real(fit_result) ** 2 + np.imag(fit_result) ** 2), '*', label="fit")
plt.legend()

plt.show()



# -*- coding: utf-8 -*-
"""
Created on Mon Jun 29 10:26:51 2015

TEST CHANGES TO FUNCTIONS CODE

Functions for the cascaded SuperSpec filter bank circuit Model. These include 
implementations of the broadband detector and spectral channel as 2 and
3-port networks. A complete list of functions is as follows:

ABCD2S - converts ABCD-matrix to S-matrix
S2ANCD - converts S-matrix to ABCD-matrix
TransmissionLine - creates a 2 or 3-port transmission line network
SpectralChannel3PortNetwork - creates a 2 or 3-port spectral channel network
FilterBank - creates a full filter bank of arbitrary size
ChannelFrequencies - calculates channel frequencies given a band and Sigma
IsolatedChannel - analytically calculates S-paramters for an isolated channel
 
@author: AlienGeorge
"""
import numpy as np
import skrf as rf
from matplotlib import pyplot as plt
import time
from multiprocessing import Pool

c = 3.0e8 # speed of light in vacuum

'''
Function to convert ABCD-matrix to S-matrix.

Arguments: A, B, C, D, Z0.
Returns: 2x2xf array S-matrix.
 
Notes: Transposition to fxnxn format is done externally.
'''
def ABCD2S(A,B,C,D, Z0):
    S11 = (1.0*A+B/Z0-C*Z0-D)/(A+B/Z0+C*Z0+D)
    S12 = 2.0*(A*D-B*C)/(A+B/Z0+C*Z0+D)
    S21 = 2.0/(A+B/Z0+C*Z0+D)
    S22 = (-1.0*A+B/Z0-C*Z0+D)/(A+B/Z0+C*Z0+D)
    return np.array([[S11, S12],[S21, S22]])
    
'''
Function to convert S-matrix to ABCD-matrix.

Arguments: S11, S12, S21, S22, Z0.
Returns 2x2xf array ABCD-matrix.

Notes: Transposition for fxnxn format is done externally.
'''
def S2ABCD(S11, S12, S21, S22, Z0):
    A = ((1.0+S11)*(1.0-S22)+S12*S21)/(2.0*S21)
    B = Z0*((1.0+S11)*(1.0+S22)-S12*S21)/(2.0*S21)
    C = 1.0/Z0*((1.0-S11)*(1.0-S22)-S12*S21)/(2.0*S21)
    D = ((1.0-S11)*(1.0+S22)+S12*S21)/(2.0*S21)
    return np.array([[A,B],[C,D]])

'''
Function to convert 2-port S-matrix with arbitrary port impedances Z01 and Z02
to Z-matrix.

Arguments: S (fx2x2 S-matrix), Z01 (fx1 Z0 vector), Z02 (fx1 Z0 vector)
Returns Z (fx2x2 Z-matrix)
'''
def S2Z(S, Z01, Z02):
    S11=S[:,0,0]; S12=S[:,0,1]; S21=S[:,1,0]; S22=S[:,1,1]
    Z11=((np.conj(Z01)+S11*Z01)*(1.0-S22)+S12*S21*Z01)/((1.0-S11)*(1.0-S22)-S12*S21)
    Z12=2.0*S12*np.sqrt(np.real(Z01)*np.real(Z02))/((1.0-S11)*(1.0-S22)-S12*S21)
    Z21=2.0*S21*np.sqrt(np.real(Z01)*np.real(Z02))/((1.0-S11)*(1.0-S22)-S12*S21)
    Z22=((1.0-S11)*(np.conj(Z02)+S22*Z02)+S12*S21*Z02)/((1.0-S11)*(1.0-S22)-S12*S21)
    Z=np.array([[Z11, Z12], [Z21, Z22]]).transpose(2, 0, 1)
    return Z
    
'''
Function to convert 2-port Z-matrix to S-matrix with arbitrary port impedances
Z01 and Z02.

Arguments: Z (fx2x2 S-matrix), Z01 (fx1 Z0 vector), Z02 (fx1 Z0 vector)
Returns S (fx2x2 S-matrix)
'''
def Z2S(Z, Z01, Z02):
    Z11=Z[:,0,0]; Z12=Z[:,0,1]; Z21=Z[:,1,0]; Z22=Z[:,1,1]
    S11=((Z11-np.conj(Z01))*(Z22+Z02)-Z12*Z21)/((Z11+Z01)*(Z22+Z02)-Z12*Z21)
    S12=2.0*Z12*np.sqrt(np.real(Z01)*np.real(Z02))/((Z11+Z01)*(Z22+Z02)-Z12*Z21)
    S21=2.0*Z21*np.sqrt(np.real(Z01)*np.real(Z02))/((Z11+Z01)*(Z22+Z02)-Z12*Z21)
    S22=((Z11+np.conj(Z01))*(Z22-np.conj(Z02))-Z12*Z21)/((Z11+Z01)*(Z22+Z02)-Z12*Z21)
    S=np.array([[S11, S12],[S21, S22]]).transpose(2, 0, 1)
    return S

'''
OLD Function to create a network object representing a transmission line.

Arguments: frequency band, physical length, char. impedance, ind. of refrac.,
attenuation constant, # of ports. 
Returns: transmission line network object
'''
def TransmissionLine(Band, l, Z0, n=1.0, alpha=0.0, nPorts=2):
    v = c/n # wave speed in transmission line    
    beta = 2.0*np.pi*Band.f/v # real propagation constant
    gamma = alpha + 1j*beta # complex propagation constant (with annuation)
    
    # construct ABCD matrix of the lossy line
    A = np.cosh(gamma*l)
    B = Z0*np.sinh(gamma*l)
    C = 1/Z0*np.sinh(gamma*l)
    D = np.cosh(gamma*l)
    
    S_2port = ABCD2S(A, B, C, D, Z0) # convert lossy line ABCD to S-parameters
    S11=S_2port[0,0]; S12=S_2port[0,1]; S21=S_2port[1,0]; S22=S_2port[1, 1];    
    
    if nPorts==2:
        S_2port = np.array([[S11, S12],[S21, S22]]).transpose(2, 0, 1)
        return rf.Network(frequency=Band, s=S_2port, z0=Z0)
    elif nPorts==3:
        S13 = np.sqrt(1-np.conj(S11)*S11-np.conj(S12)*S12)    
        S31 = S13
        theta23 = np.pi/2-beta*l    
        S23 = np.abs(S13)*np.exp(1j*theta23)
        S32 = S23
        S33 = np.zeros(np.size(Band.f))
        S_3port = np.array([[S11, S12, S13],[S21, S22, S23],[S31, S32, S33]]).transpose(2, 0, 1)
        return rf.Network(frequency = Band, s = S_3port, z0 = Z0)

'''
Function to create a network object representing a lossy transmission line.

Arguments: frequency band, physical length, char. impedance, propagation speed,
relative permittivity of dielectric, loss tangent of dielectric, # of ports. 
Returns: transmission line network object
Note: we assume that loss is dominated by dielectric loss
'''
def TransmissionLineLossy(Band, l, Z0, v, epsr=11.7, lossTan=0.0, nPorts=2):   
    beta = 2.0*np.pi*Band.f/v # real propagation constant
    #alpha = np.pi*np.sqrt(epsr)*Band.f/v*lossTan # attenuation constant in Np/m
    alpha = np.pi*Band.f/v*lossTan # attenuation constant in Np/m so that the units are on chip normalized for epsilonr
    gamma = alpha + 1j*beta # complex propagation constant (with attenuation)
    
    # construct ABCD matrix of the lossy line
    A = np.cosh(gamma*l)
    B = Z0*np.sinh(gamma*l)
    C = 1/Z0*np.sinh(gamma*l)
    D = np.cosh(gamma*l)
    
    S_2port = ABCD2S(A, B, C, D, Z0) # convert lossy line ABCD to S-parameters
    S11=S_2port[0,0]; S12=S_2port[0,1]; S21=S_2port[1,0]; S22=S_2port[1, 1];    
    
    if nPorts==2:
        S_2port = np.array([[S11, S12],[S21, S22]]).transpose(2, 0, 1)
        return rf.Network(frequency=Band, s=S_2port, z0=Z0)
    elif nPorts==3:
        S13 = np.sqrt(1-np.conj(S11)*S11-np.conj(S12)*S12)    
        S31 = S13
        theta23 = np.pi/2-beta*l    
        S23 = np.abs(S13)*np.exp(1j*theta23)
        S32 = S23
        S33 = np.zeros(np.size(Band.f))
        S_3port = np.array([[S11, S12, S13],[S21, S22, S23],[S31, S32, S33]]).transpose(2, 0, 1)
        return rf.Network(frequency = Band, s = S_3port, z0 = Z0)

'''
OLD Function to create a network object representing a spectral channel as
a 3-port network.

Arguments: freq. band, char. impedance, resonant freq., coupling Q, internal Q,
approach ("1" for matched ports, "2" for unmatched ports).
Returns: spectral channel network object
'''
def SpectralChannel3PortNetwork(Band, Z0, fres, Qc, Qi, approach=1):
    x = (Band.f-fres)/fres
    ZL = Z0/2*Qc/Qi+1j*Z0*Qc*x   # shunt impedance of the entire resonator
    # ZR = ZL-Z0  # resonator impedance minus impedance of matched port
    
    # 3-port network with all ports referenced to 50 Ohm; terminate port 2 with ZL
    if approach==1:
        S_3port = np.ones((np.size(x), 3, 3))
        
        # create 3-port S-matrix for the 50 Ohm matched network
        S_3port[:,0, 0]=-1.0/3; S_3port[:,0, 1]=2.0/3; S_3port[:,0, 2]=2.0/3
        S_3port[:,1, 0]=2.0/3; S_3port[:,1, 1]=-1.0/3; S_3port[:,1, 2]=2.0/3
        S_3port[:,2, 0]=2.0/3; S_3port[:,2, 1]=2.0/3; S_3port[:,2, 2]=-1.0/3
        
        # create a network object for the 50 Ohm matched network        
        MatchedNtwk = rf.Network(frequency = Band, s = S_3port, z0 = Z0)
        
        # create 1-port S-matrix for the resonator load
        S11_L=(ZL-Z0)/(ZL+Z0)
        Resonator = rf.Network(frequency = Band, s = S11_L, z0 = Z0)
        Ntwk = rf.connect(MatchedNtwk, 2, Resonator, 0)
        return Ntwk
    
    # 3-port network with all ports referenced to 50 Ohm; renormalize so port 2
    # is referenced to ZL
    elif approach==2:
        S_3port = np.ones((np.size(x), 3, 3))
        
        # create 3-port S-matrix for the 50 Ohm matched network
        S_3port[:,0, 0]=-1.0/3; S_3port[:,0, 1]=2.0/3; S_3port[:,0, 2]=2.0/3
        S_3port[:,1, 0]=2.0/3; S_3port[:,1, 1]=-1.0/3; S_3port[:,1, 2]=2.0/3
        S_3port[:,2, 0]=2.0/3; S_3port[:,2, 1]=2.0/3; S_3port[:,2, 2]=-1.0/3
        
        # create a network object for the 50 Ohm matched network        
        Ntwk = rf.Network(frequency = Band, s = S_3port, z0 = Z0)
        
        # create port reference impedance matrix
        Zmatrix = np.empty([len(Band.f),3], dtype=complex)
        Zmatrix[:,2]=Z0; Zmatrix[:,0]=Z0; Zmatrix[:,1]=ZL
        
        Ntwk.renormalize(Zmatrix, powerwave=True)
        return Ntwk
#    elif approach==2:
#        # create 3-port S-matrix for the network referenced at ZL for port 3        
#        S11=-Z0/(2*ZL+Z0); S22=S11; S12=2*ZL/(2*ZL+Z0); S21=S12
#        S31=np.sqrt(Z0/np.real(ZL))*2*np.real(ZL)/(2*ZL+Z0); S32=S31
#        S13=np.sqrt(np.real(ZL)/Z0)*2*(Z0+4*np.imag(ZL))/(2*ZL+Z0); S23=S13
#        S33=(Z0-ZL)/(Z0+2*np.conj(ZL))
#        S_3port = np.array([[S11, S12, S13],[S21, S22, S23],[S31, S32, S33]]).transpose(2,0,1)
#        
#        # create port reference impedance matrix
#        Zmatrix = np.empty([len(x),3], dtype=complex)
#        Zmatrix[:,2]=Z0; Zmatrix[:,0]=Z0; Zmatrix[:,1]=Z0
#        
#        return rf.Network(frequency = Band, s = S_3port, z0=Zmatrix)
    
    elif approach==3:
        # create 2-port S-matrix for the network
        S11=-Z0/(2*ZL+Z0); S22=S11; S12=2*ZL/(2*ZL+Z0); S21=S12
        S_2port=np.array([[S11, S12],[S21, S22]]).transpose(2, 0, 1)
        
        return rf.Network(frequency = Band, s = S_2port, z0=Z0)

'''
Function to create a network object representing a spectral channel as either a
2-port or 3-port network. Incorporates lossy dielectric as Qloss.

Arguments: freq. band, char. impedance, resonant freq., coupling Q, internal Q,
loss Q, approach ("1" for 3-port representation all referenced to Z0 and
terminated in ZL, "2" for 3-port representation with port 2 referenced to ZL, 
"3" for 2-port representation).
Returns: spectral channel network object
'''
def SpectralChannelLossy(Band, Z0, fres, Qc, Qdet, Qloss, approach=1,normalize = True):
    x = (Band.f-fres)/fres
    ZL = Z0/2*Qc*(1/Qdet+1/Qloss)+1j*Z0*Qc*x   # shunt impedance of the entire resonator
    # ZR = ZL-Z0  # resonator impedance minus impedance of matched port
    
    # 3-port network with all ports referenced to 50 Ohm; terminate port 2 with ZL
    if approach==1:
        S_3port = np.ones((np.size(x), 3, 3))
        
        # create 3-port S-matrix for the 50 Ohm matched network
        S_3port[:,0, 0]=-1.0/3; S_3port[:,0, 1]=2.0/3; S_3port[:,0, 2]=2.0/3
        S_3port[:,1, 0]=2.0/3; S_3port[:,1, 1]=-1.0/3; S_3port[:,1, 2]=2.0/3
        S_3port[:,2, 0]=2.0/3; S_3port[:,2, 1]=2.0/3; S_3port[:,2, 2]=-1.0/3
        
        # create a network object for the 50 Ohm matched network        
        MatchedNtwk = rf.Network(frequency = Band, s = S_3port, z0 = Z0)
        
        # create 1-port S-matrix for the resonator load
        S11_L=(ZL-Z0)/(ZL+Z0)
        Resonator = rf.Network(frequency = Band, s = S11_L, z0 = Z0)
        Ntwk = rf.connect(MatchedNtwk, 2, Resonator, 0)
        return Ntwk
    
    # 3-port network with all ports referenced to 50 Ohm; renormalize so port 2
    # is referenced to ZL
    elif approach==2:
        S_3port = np.ones((np.size(x), 3, 3))
        
        # create 3-port S-matrix for the 50 Ohm matched network
        # had to add in small values to allow matrix to be imported jordan 5/29/2020
        S_3port[:,0, 0]=-1.0/3+0.0000001; S_3port[:,0, 1]=2.0/3; S_3port[:,0, 2]=2.0/3
        S_3port[:,1, 0]=2.0/3; S_3port[:,1, 1]=-1.0/3+0.0000001; S_3port[:,1, 2]=2.0/3
        S_3port[:,2, 0]=2.0/3; S_3port[:,2, 1]=2.0/3; S_3port[:,2, 2]=-1.0/3+0.0000001
        #print(S_3port)
        # create a network object for the 50 Ohm matched network        
        Ntwk = rf.Network(frequency = Band, s = S_3port, z0 = Z0)
        #print(Ntwk)
        #print(Ntwk.z0)
        #print(Ntwk.z0.shape)
        #print(Ntwk.s)
        #print(Ntwk.s.shape)
        # create port reference impedance matrix
        if normalize:
            Zmatrix = np.empty([len(Band.f),3], dtype=complex)
            Zmatrix[:,2]=Z0
            Zmatrix[:,0]=Z0
            Zmatrix[:,1]=ZL
            #print("ZL is",ZL)
            #print("Z0 is",Z0)
            #print(Zmatrix)
            #print(Zmatrix.shape)
            Ntwk.renormalize(Zmatrix,s_def='power')#, powerwave=True)
        return Ntwk
    
    elif approach==3:
        # create 2-port S-matrix for the network
        S11=-Z0/(2*ZL+Z0); S22=S11; S12=2*ZL/(2*ZL+Z0); S21=S12
        S_2port=np.array([[S11, S12],[S21, S22]]).transpose(2, 0, 1)
        
        return rf.Network(frequency = Band, s = S_2port, z0=Z0)
    
'''
OLD Function to create a network object for a filter bank with an arbitrary # of 
channels, reorder the ports to the standard filter bank convention, and 
optionally plot the thru power and power absorbed by each channel. User input 
determines whether or not to plot, plot title, start/stop frequency, whether or
not to save, and file name.

Arguments: freq. band, design data for all channels, char. impedance, ind. of 
refrac., physical separation between channels. 
Returns: filter bank network object with correctly ordered ports

Notes: Physical separation refers to the wavelength corresponding to the resonant
frequency of left channel in each pair of channels.
'''
def FilterBank(Band, Data, Z0=50.0, n=1.0, physSep = 0.25, approach=3, doPlot=False):
    v = c/n
    # initialize current network to the first spectral channel
    CurrentNtwk = SpectralChannel3PortNetwork(Band, Z0, Data[0,0], Data[1,0], Data[2,0], approach)
        
    # loop to create filter bank with arbitrary # of channels and create network
    for i in range(np.shape(Data)[1]):
        if i < np.shape(Data)[1]-1:
            # resonant frequencies and quality factors for current and next SCs
            fres_current = Data[0,i]    
            fres_nxt = Data[0,i+1]; Qc_nxt = Data[1,i+1]; Qi_nxt = Data[2,i+1]
            
            # create Network object for next SC
            NextSC = SpectralChannel3PortNetwork(Band, Z0, fres_nxt, Qc_nxt, Qi_nxt, approach)
            
            # create interconnecting transmission line
            lambda_current = v/fres_current
            # lambda_nxt = v/fres_nxt
            lineLength = physSep*lambda_current
            # lineLength = physSep*(lambda_current+lambda_nxt)/2.0
            TLine = TransmissionLine(Band, lineLength, Z0, n)
            
            # connect current network to the transmission line
            N = CurrentNtwk.nports
            InterNtwk = rf.connect(CurrentNtwk, N-1, TLine, 0)
            
            # connect current network to the next SC
            N = InterNtwk.nports
            CurrentNtwk = rf.connect(InterNtwk, N-1, NextSC, 0)
    
    # for 3-port approach, reorder port numbers so S10 is thru & channels are in order from 2 to N-1
#    if approach==2:
#        N = CurrentNtwk.nports    
#        OldS = CurrentNtwk.s
#        NewS = np.empty([len(Band), N, N], dtype=complex)
#        NewS[:,0,0]=OldS[:,0,0]; NewS[:,0,N-1]=OldS[:,0,N-1]
#        NewS[:,N-1,0]=OldS[:,N-1,0]; NewS[:,N-1,N-1]=OldS[:,N-1,N-1];
#        
#        # ugly numerology code
#        for l in range(N):
#            for k in range(N):
#                if l==0 or l==N-1:
#                    if k in range(1,N-2):
#                        NewS[:,k+1,l]=OldS[:,k,l]
#                    elif k==N-2:
#                        NewS[:,1,l]=OldS[:,k,l]
#                elif l==N-2:
#                    if k in range(1,N-2):
#                        NewS[:,k+1,1]=OldS[:,k,l]
#                    elif k==0:
#                        NewS[:,k,1]=OldS[:,k,l]
#                    elif k==N-2:
#                        NewS[:,1,1]=OldS[:,k,l]
#                elif l in range(1,N-2):
#                    if k in range(1,N-2):
#                        NewS[:,k+1,l+1]=OldS[:,k,l]
#                    elif k==0:
#                        NewS[:,k,l+1]=OldS[:,k,l]
#                    elif k==N-2:
#                        NewS[:,1,l+1]=OldS[:,k,l]
#                    elif k==N-1:
#                        NewS[:,k,l+1]=OldS[:,k,l]
#                        
#        # create a new network with the correct port order to represent the filter bank
#        CurrentNtwk = rf.Network(frequency = Band, s = NewS, z0 = CurrentNtwk.z0)     
    
    # code to do plotting if doPlot is True      
    if doPlot:
        # full filter bank S-Matrix with correct port ordering
        nPorts = CurrentNtwk.number_of_ports
        
        # readout and plot reflected, thru, and each individual channel
#        PowerS = np.empty([len(Band), nPorts], dtype=complex)
#        PowerS[:,0] = CurrentNtwk.s[:,0,0]; PowerS[:,1] = CurrentNtwk.s[:,nPorts-2,0]
        
        # plots for arbitrary number of channels
        plt.figure(1); plt.clf()
        plt.rc('legend', fontsize = 12)
#        plt.plot(Band.f/1.0e9, np.conj(PowerS[:,1])*PowerS[:,1], linewidth = '2', 
#                 label = 'Thru')
        plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,1,0])*CurrentNtwk.s[:,1,0], linewidth = '2', 
                 label = 'Thru')
        plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,0,0])*CurrentNtwk.s[:,0,0], linewidth = '2', 
                 label = 'Reflected')
        
        # plot absorbed powers if using 3-port approach for spectral channels
        if approach == 2:
            # if there are 5 or less channels, each channel is different color
            if np.shape(Data)[1] <= 5:        
                for i in range(2, nPorts-1):
                    # PowerS[:,i]=CurrentNtwk.s[:,i-1,0]
                    plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,i,0])*CurrentNtwk.s[:,i,0], '--', linewidth = '2', 
                             label = 'Ch ' + str(i-1))
                # PowerS[:,nPorts-1] = CurrentNtwk.s[:,nPorts-1,0]
                plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,nPorts-1,0])*CurrentNtwk.s[:,nPorts-1,0], '--', linewidth = '2', 
                         label = 'Ch ' + str(nPorts-2))
                         
            # otherwise plot all channels in black with no legend entries
            else:
                for i in range(2, nPorts-1):
                    # PowerS[:,i]=CurrentNtwk.s[:,i-1,0]
                    plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,i,0])*CurrentNtwk.s[:,i,0], 'k:', linewidth = '3')
    
                #PowerS[:,nPorts-1] = CurrentNtwk.s[:,nPorts-1,0]
                plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,nPorts-1,0])*CurrentNtwk.s[:,nPorts-1,0], 'k:', linewidth = '3')
        
        # prompt user to enter plot title and start/stop frequencies
        plotTitle = raw_input('What would you like to title your plot? ')
        start = raw_input('Start frequency (GHz): ')
        stop = raw_input('Stop frequency (GHz): ')
        
        plt.title(plotTitle)
        plt.legend(loc='best')
        plt.xlabel('Frequency (GHz)')
        plt.ylabel('Power')
        plt.xlim(float(start), float(stop))
        plt.ylim(0, 1)
        plt.ion()
        plt.show()
        
        # prompt user about whether or not to save plot
        while True:
            isSave = raw_input('Would you like to save the plot? Y or N ')
            if isSave in ('y', 'n', 'Y', 'N'):
                break
            else:
                print("""You must enter 'Y' or 'N'! """)
        
        # if yes, prompt for file name and save as .png file
        if isSave in ('y', 'Y'):
            plotFileName = raw_input('What name would you like to save your plot as? ')
            plt.savefig('PythonPlots/' + plotFileName + '.png') 
    return CurrentNtwk

'''
Function to create an unterminated network object for a filter bank with an 
arbitrary # of channels taking into account possible dielectric loss and 
optionally plot the reflected power, thru power, and power absorbed by each 
channel. User input determines whether or not to plot, plot title, start/stop 
frequency, whether or not to save, and file name.

Arguments: freq. band, design data for all channels, char. impedance, 
propagation speed on feedline, physical separation between channels (in 
wavelengths), transmission line attenuation, spectral channel approach, to plot
or not to plot. 
Returns: filter bank network object

Notes: Physical separation refers to the wavelength corresponding to the resonant
frequency of left channel in each pair of channels.
'''
def FilterBankLossy(Band, Data, Z0=50.0, v=c, physSep=0.25, epsr=11.7, lossTan=0.0, approach=3,processors = 1, doPlot=False):
    
    # initialize current network to the first spectral channel
    net_dict = {}
    CurrentNtwk = SpectralChannelLossy(Band, Z0, Data[0,0], Data[1,0], Data[2,0], Data[3,0], approach)
    net_dict['net0'] = CurrentNtwk
    then = time.time()    
    # loop to create filter bank with arbitrary # of channels and create network
    for i in range(np.shape(Data)[1]):
        if i < np.shape(Data)[1]-1:
            # resonant frequencies and quality factors for current and next SCs
            fres_current = Data[0,i]    
            fres_nxt = Data[0,i+1]; Qc_nxt = Data[1,i+1]; Qdet_nxt = Data[2,i+1]; Qloss_nxt = Data[3,i+1]
            
            # create Network object for next SC
            #NextSC = SpectralChannelLossy(Band, Z0, fres_nxt, Qc_nxt, Qdet_nxt, Qloss_nxt, approach)
            NextSC = connect_spectral_and_Tline(Band, Z0, fres_nxt, Qc_nxt, Qdet_nxt, Qloss_nxt, approach,v,physSep,epsr)
            net_dict['net'+str(i+1)] = NextSC
            # connect current network to the transmission line
            #N = CurrentNtwk.nports
            #CurrentNtwk = rf.connect(CurrentNtwk, N-1, NextSC, 0)
    print(time.time() -then)
    #print(len(net_dict.keys()))
    then = time.time()                    
    fast_net = connect_net_dict_multi(net_dict,processors = processors)
    #CurrentNtwk = fast_net[0] #uncomment for parrallel processsing
    #print(fast_net.keys())
    CurrentNtwk = fast_net['net0']
    #print(CurrentNtwk.nports)
    #print(CurrentNtwk.f)
    print(time.time() -then)
    then = time.time()    
    # renormalize the ports out here becuase it is faster that doing for each spectral channel individually
    Zmatrix = np.empty([len(Band.f),CurrentNtwk.nports], dtype=complex)
    Zmatrix[:,CurrentNtwk.nports-1]=Z0
    Zmatrix[:,0]=Z0
    for i in range(1,CurrentNtwk.nports-1):
        x = (Band.f-Data[0,i-1])/Data[0,i-1]
        ZL = Z0/2*Data[1,i-1]*(1/Data[2,i-1]+1/Data[3,i-1])+1j*Z0*Data[1,i-1]*x
        Zmatrix[:,i]=ZL
    print(Zmatrix.shape)
    CurrentNtwk.renormalize(Zmatrix,s_def='power')#, powerwave=True)
    print(time.time() -then)
            
    # code to do plotting if doPlot is True      
    if doPlot:
        # full filter bank S-Matrix with correct port ordering
        nPorts = CurrentNtwk.number_of_ports
        
        # plots for arbitrary number of channels
        plt.figure(1); plt.clf()
        plt.rc('legend', fontsize = 12)

        plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,nPorts-1,0])*CurrentNtwk.s[:,nPorts-1,0], linewidth = '2', 
                 label = 'Thru')
        plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,0,0])*CurrentNtwk.s[:,0,0], linewidth = '2', 
                 label = 'Reflected')
        
        # plot absorbed powers if using 3-port approach for spectral channels
        if approach == 2:
            # if there are 5 or less channels, each channel is different color
            if np.shape(Data)[1] <= 5:        
                for i in range(1, nPorts-1):
                    # PowerS[:,i]=CurrentNtwk.s[:,i-1,0]
                    plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,i,0])*CurrentNtwk.s[:,i,0], '--', linewidth = '2', 
                             label = 'Ch ' + str(i-1))
                # PowerS[:,nPorts-1] = CurrentNtwk.s[:,nPorts-1,0]
                plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,nPorts-1,0])*CurrentNtwk.s[:,nPorts-1,0], '--', linewidth = '2', 
                         label = 'Ch ' + str(nPorts-2))
                         
            # otherwise plot all channels in black with no legend entries
            else:
                for i in range(1, nPorts-1):
                    # PowerS[:,i]=CurrentNtwk.s[:,i-1,0]
                    plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,i,0])*CurrentNtwk.s[:,i,0], 'k:', linewidth = '3')
    
                #PowerS[:,nPorts-1] = CurrentNtwk.s[:,nPorts-1,0]
                plt.plot(Band.f/1.0e9, np.conj(CurrentNtwk.s[:,nPorts-1,0])*CurrentNtwk.s[:,nPorts-1,0], 'k:', linewidth = '3')
        
        # prompt user to enter plot title and start/stop frequencies
        plotTitle = raw_input('What would you like to title your plot? ')
        start = raw_input('Start frequency (GHz): ')
        stop = raw_input('Stop frequency (GHz): ')
        
        plt.title(plotTitle)
        plt.legend(loc='best')
        plt.xlabel('Frequency (GHz)')
        plt.ylabel('Power')
        plt.xlim(float(start), float(stop))
        plt.ylim(0, 1)
        plt.ion()
        plt.show()
        
        # prompt user about whether or not to save plot
        while True:
            isSave = raw_input('Would you like to save the plot? Y or N ')
            if isSave in ('y', 'n', 'Y', 'N'):
                break
            else:
                print("""You must enter 'Y' or 'N'! """)
        
        # if yes, prompt for file name and save as .png file
        if isSave in ('y', 'Y'):
            plotFileName = raw_input('What name would you like to save your plot as? ')
            plt.savefig('PythonPlots/' + plotFileName + '.png') 
    return CurrentNtwk

'''
Function to create a network object for a filter bank terminated in an arbitrary
load. Also connects length of transmission line between the final channel and
termination.

Arguments: same as for FilterBank + length of line preceding filter bank, length of line after filter bank, impedance of
termination

Returns: terminated filter bank network object
'''
def TerminatedFilterBank(Band, Data, lBack, ZT, Z0=50.0, n=1.0, physSep=0.25, approachFB=3, approachT=1):
    UnterminatedFB = FilterBank(Band, Data, Z0, n, physSep, approachFB)
    
    # transmission line between final channel and termination
    TLine = TransmissionLine(Band, lBack, Z0, n)

    # convert Z-matrix to S-matrix with port 2 referenced to termination impedance
    # resulting in a 2-port Ntwk; used for computing response of BB detector after
    # the spectral channels; only works if SC is represented by 2-port network
    if approachT==1:
        # 2-port network with FB connected to end transmission line
        InterNtwk = rf.connect(UnterminatedFB, 1, TLine, 0)
        Z01=Z0*np.ones(Band.npoints) # InterNtwk port 1 ref. impedance
        Z02=ZT*np.ones(Band.npoints) # InterNtwk port 2 ref. impedance
        Z0Matrix=np.vstack((Z01, Z02)).transpose(1,0)
        TermS=Z2S(InterNtwk.z, Z01, Z02)
        Ntwk=rf.Network(frequency=Band, s=TermS, z0=Z0Matrix)
        
    # create 1-port network for termination and connect to unterminated FB resulting
    # in a 1-port Ntwk
    elif approachT==2:
        # create 1-port S-matrix for the termination
        S11_T=(ZT*np.ones(Band.npoints)-Z0)/(ZT*np.ones(Band.npoints)+Z0)
        Term=rf.Network(frequency=Band, s=S11_T, z0=Z0)
        
        # 3-port representation of spectral channels
        if approachFB==2:
            InterNtwk = rf.connect(UnterminatedFB, UnterminatedFB.nports-1, TLine, 0)
            # terminate filter bank
            Ntwk = rf.connect(InterNtwk, InterNtwk.nports-1, Term, 0)
            
        # 2-port representation of spectral channels
        elif approachFB==3:
            InterNtwk = rf.connect(UnterminatedFB, 1, TLine, 0)
            # terminate filter bank
            Ntwk = rf.connect(InterNtwk, 1, Term, 0)
    
    return Ntwk

'''
Function to create a network object for a filter bank terminated in an arbitrary
load taking in account possible dielectric loss. Also connects length of 
transmission line between the final channel and termination.

Arguments: same as for FilterBankLossy + 
length of line after filter bank, impedance of termination

Returns: terminated filter bank network object
'''
def TerminatedFilterBankLossy(Band, Data, lBack, ZT, Z0=50.0, v=c, physSep=0.25, epsr=11.7, lossTan=0.0, approachFB=3, approachT=1):
    UnterminatedFB = FilterBankLossy(Band, Data, Z0, v, physSep, epsr, lossTan, approachFB)

    # transmission line between final channel and termination
    TLine = TransmissionLineLossy(Band, lBack, Z0, v, epsr, lossTan)

    # convert Z-matrix to S-matrix with port 2 referenced to termination impedance
    # resulting in a 2-port Ntwk; used for computing response of BB detector after
    # the spectral channels; only works if SC is represented by 2-port network
    if approachT==1:
        # 2-port network with FB connected to end transmission line
        InterNtwk = rf.connect(UnterminatedFB, 1, TLine, 0)
        Z01=Z0*np.ones(Band.npoints) # InterNtwk port 1 ref. impedance
        Z02=ZT*np.ones(Band.npoints) # InterNtwk port 2 ref. impedance
        Z0Matrix=np.vstack((Z01, Z02)).transpose(1,0)
        TermS=Z2S(InterNtwk.z, Z01, Z02)
        Ntwk=rf.Network(frequency=Band, s=TermS, z0=Z0Matrix)
        
    # create 1-port network for termination and connect to unterminated FB resulting
    # in a 1-port Ntwk
    elif approachT==2:
        # create 1-port S-matrix for the termination
        S11_T=(ZT*np.ones(Band.npoints)-Z0)/(ZT*np.ones(Band.npoints)+Z0)
        Term=rf.Network(frequency=Band, s=S11_T, z0=Z0)
        
        # 3-port representation of spectral channels
        if approachFB==2:
            InterNtwk = rf.connect(UnterminatedFB, UnterminatedFB.nports-1, TLine, 0)
            # terminate filter bank
            Ntwk = rf.connect(InterNtwk, InterNtwk.nports-1, Term, 0)
            
        # 2-port representation of spectral channels
        elif approachFB==3:
            InterNtwk = rf.connect(UnterminatedFB, 1, TLine, 0)
            # terminate filter bank
            Ntwk = rf.connect(InterNtwk, 1, Term, 0)
    
    return Ntwk

'''
Function to create a network object for a filter bank terminated in an arbitrary
load taking in account possible dielectric loss. Also connects length of 
transmission line between the final channel and termination. Used to
calculate response of BB1, a broadband absorber placed before the spectral channels.

Arguments: same as for FilterBankLossy + 
length of line after filter bank, impedance of termination

Returns: terminated filter bank network object
'''
def TerminatedFBLossyEnd(Band, Data, lBack, ZT, Z0=50.0, v=c, physSep=0.25, epsr=11.7, lossTan=0.0, approachFB=3, approachT=1):
    UnterminatedFB = FilterBankLossy(Band, Data, Z0, v, physSep, epsr, lossTan, approachFB)

    # transmission line between final channel and termination
    TLine = TransmissionLineLossy(Band, lBack, Z0, v, epsr, lossTan)

    # convert Z-matrix to S-matrix with port 2 referenced to termination impedance
    # resulting in a 2-port Ntwk; used for computing response of BB detector after
    # the spectral channels; only works if SC is represented by 2-port network
    if approachT==1:
        # 2-port network with FB connected to end transmission line
        InterNtwk = rf.connect(UnterminatedFB, 1, TLine, 0)
        Z01=Z0*np.ones(Band.npoints) # InterNtwk port 1 ref. impedance
        Z02=ZT*np.ones(Band.npoints) # InterNtwk port 2 ref. impedance
        Z0Matrix=np.vstack((Z01, Z02)).transpose(1,0)
        TermS=Z2S(InterNtwk.z, Z01, Z02)
        Ntwk=rf.Network(frequency=Band, s=TermS, z0=Z0Matrix)
        
    # create 1-port network for termination and connect to unterminated FB resulting
    # in a 1-port Ntwk
    elif approachT==2:
        # create 1-port S-matrix for the termination
        S11_T=(ZT*np.ones(Band.npoints)-Z0)/(ZT*np.ones(Band.npoints)+Z0)
        Term=rf.Network(frequency=Band, s=S11_T, z0=Z0)
        
        # 3-port representation of spectral channels
        if approachFB==2:
            InterNtwk = rf.connect(UnterminatedFB, UnterminatedFB.nports-1, TLine, 0)
            # terminate filter bank
            Ntwk = rf.connect(InterNtwk, InterNtwk.nports-1, Term, 0)
            
        # 2-port representation of spectral channels
        elif approachFB==3:
            InterNtwk = rf.connect(UnterminatedFB, 1, TLine, 0)
            # terminate filter bank
            Ntwk = rf.connect(InterNtwk, 1, Term, 0)
            
    # same as approachT = 1, but using renormalize function instead of converting
    # between S and Z parameters
    elif approachT==3:
        Ntwk = rf.connect(UnterminatedFB, UnterminatedFB.nports-1, TLine, 0)
            
        NewPortZ = np.empty([Band.npoints, Ntwk.nports], dtype=complex)
        NewPortZ[:,Ntwk.nports-1]=ZT
        for i in range(0, Ntwk.nports-1):
            NewPortZ[:,i] = Ntwk.z0[:,i]
        Ntwk.renormalize(NewPortZ)
    return Ntwk

'''
Function to create a network object for a filter bank preceeded by an antenna
with arbitrary impedance taking into account possible dielectric loss. Also connects 
length of transmission line between the antenna and first channel. Used to
calculate response of BB2, a broadband absorber placed after the spectral channels.

Arguments: same as for FilterBankLossy + length of line preceding filter bank, 
impedance of antenna

Returns: terminated filter bank network object
'''
def TerminatedFBLossyBegin(Band, Data, lFront, ZA, Z0=50.0, v=c, physSep=0.25, epsr=11.7, lossTan=0.0, approachFB=3, approachT=1):
    UnterminatedFB = FilterBankLossy(Band, Data, Z0, v, physSep, epsr, lossTan, approachFB)

    # transmission line between antenna and first channel
    TLine = TransmissionLineLossy(Band, lFront, Z0, v, epsr, lossTan)

    # convert Z-matrix to S-matrix with port 2 referenced to termination impedance
    # resulting in a 2-port Ntwk; used for computing response of BB detector after
    # the spectral channels; only works if SC is represented by 2-port network
    if approachT==1:
        # 2-port network with FB connected to end transmission line
        InterNtwk = rf.connect(TLine, 1, UnterminatedFB, 0)
        Z01=ZA*np.ones(Band.npoints) # InterNtwk port 1 ref. impedance
        Z02=Z0*np.ones(Band.npoints) # InterNtwk port 2 ref. impedance
        Z0Matrix=np.vstack((Z01, Z02)).transpose(1,0)
        TermS=Z2S(InterNtwk.z, Z01, Z02)
        Ntwk=rf.Network(frequency=Band, s=TermS, z0=Z0Matrix)
        
    # same as approachT = 1, but using renormalize instead of converting between
    # S and Z parameters. If using 3-port SC representation, the through S-parameter
    # is S[nports-1,0]
    elif approachT==2:
        # for 2-port and 3-port representations of spectral channels
        Ntwk = rf.connect(TLine, 1, UnterminatedFB, 0)
        NewPortZ = np.empty([Band.npoints, Ntwk.nports], dtype=complex)
        NewPortZ[:,0]=ZA
        for i in range(1, Ntwk.nports):
            NewPortZ[:,i] = Ntwk.z0[:,i]
        Ntwk.renormalize(NewPortZ)
    return Ntwk

'''
Function to create a network object for a filter bank preceeded by an antenna
with arbitrary impedance AND terminated in an arbitrary load taking into account 
possible dielectric loss. Lengths of transmission line connect antenna to first
channel and final channel to termination. Used to calculate response of each 
spectral channel under realistic input and output port termination conditions.

Arguments: same as for FilterBankLossy + length of line preceding filter bank, 
impedance of antenna, length of line following filter bank, impedance of
termination

Returns: terminated filter bank network object
'''
def TerminatedFBLossyBoth(Band, Data, lFront, lBack, ZA, ZT, Z0=50.0, v=c, physSep=0.25, epsr=11.7, lossTan=0.0, approachFB=3, approachT=1,processors = 1):
    
    UnterminatedFB = FilterBankLossy(Band, Data, Z0, v, physSep, epsr, lossTan, approachFB, processors)
    

    # transmission line between antenna and first channel
    TLineBegin = TransmissionLineLossy(Band, lFront, Z0, v, epsr, lossTan)
    # transmission line between final channel and termination
    TLineEnd = TransmissionLineLossy(Band, lBack, Z0, v, epsr, lossTan)
    
    # convert Z-matrix to S-matrix with port 2 referenced to termination impedance
    # resulting in a 2-port Ntwk; only works if SC is represented by 2-port network
    if approachT==1:
        # 2-port network with FB connected to end transmission line
        InterNtwk1 = rf.connect(TLineBegin, 1, UnterminatedFB, 0)
        InterNtwk2 = rf.connect(InterNtwk1, 1, TLineEnd, 0)
        Z01=ZA*np.ones(Band.npoints) # InterNtwk port 1 ref. impedance
        Z02=ZT*np.ones(Band.npoints) # InterNtwk port 2 ref. impedance
        Z0Matrix=np.vstack((Z01, Z02)).transpose(1,0)
        TermS=Z2S(InterNtwk2.z, Z01, Z02)
        Ntwk=rf.Network(frequency=Band, s=TermS, z0=Z0Matrix)
        
    # same as approachT = 1, but using renormalize instead of converting between
    # S and Z parameters. If using 3-port SC representation, the through S-parameter
    # is S[nports-1,0]
    elif approachT==2:
        # for 2-port and 3-port representations of spectral channels
        InterNtwk = rf.connect(TLineBegin, 1, UnterminatedFB, 0)
        Ntwk = rf.connect(InterNtwk, InterNtwk.nports-1, TLineEnd, 0)
        
        NewPortZ = np.empty([Band.npoints, Ntwk.nports], dtype=complex)
        NewPortZ[:,0]=ZA
        NewPortZ[:,Ntwk.nports-1]=ZT
        for i in range(1, Ntwk.nports-1):
            NewPortZ[:,i] = Ntwk.z0[:,i]
        Ntwk.renormalize(NewPortZ)
    
    return Ntwk

'''
Function to calculate response of BB1, a broadband absorber placed inline with 
the transmission line in front of a terminated filter bank.

Arguments: coupling constant of BB1, frequency band, transmission line length, 
coupling length, center position, characteristic impedance of transmission line, 
antenna impedance, propagation speed on transmission line, dielectric constant, 
loss tangent of dielectric layer, reflection coefficient at input of terminated 
filter bank (assuming matched impedance), approach (1 normalizes response to V 
at antenna, 2 normalizes response to V after antenna)

Returns: an array of BB1 response at each frequency point in the band
'''
def BroadbandChannelBefore(epsilon1, Band, lLine, lCoupling, zc, Z0, ZA, v, epsr, lossTan, GammaFB, approach=1):
    beta = 2.0*np.pi*Band.f/v # real propogation constant
    alpha = np.pi*np.sqrt(epsr)*Band.f/v*lossTan # attenuation constant in Np/m
    gamma = alpha + 1j*beta # complex propagation constant (with attenuation)
    
    # input impedance immediately after antenna looking toward termination
    Zin = Z0*(1.0+GammaFB*np.exp(-2.0*gamma*lLine))/(1.0-GammaFB*np.exp(-2.0*gamma*lLine))
    # incident voltage at end of transmission line
    V0plus = Zin/((ZA+Zin)*(np.exp(gamma*lLine)+GammaFB*np.exp(-gamma*lLine)))
    
    # power coupled to BB1 normalized to power immediately after antenna
    BB1 = epsilon1*(np.sinc(1j*alpha*lCoupling)*np.exp(-2*alpha*zc)\
                                   + 2.0*np.abs(GammaFB)*np.cos(2.0*beta*zc+np.angle(GammaFB))\
                                   *np.sinc(beta*lCoupling) + np.conj(GammaFB)*GammaFB\
                                   *np.sinc(1j*alpha*lCoupling)*np.exp(2.0*alpha*zc))
    
    # unity is Voc at antenna
    if approach==1:
        return np.conj(V0plus)*V0plus*BB1
    # unity is at end of transmission line, immediately before the first channel
    elif approach==2:
        return BB1
    
'''
Function to calculate response of BB2, a broadband absorber placed inline with 
the transmission line behind of a terminated filter bank.

Arguments: coupling constant of BB2, frequency band, coupling length, center 
position, propagation speed on transmission line, dielectric constant, 
loss tangent of dielectric layer, transmission coefficient (S21) at termination
of filter bank (filter bank object used must contain end transmission line with
load port referenced to termination impedance)

Returns: an array of BB2 response at each frequency point in the band
'''
def BroadbandChannelAfter(epsilon2, Band, lCoupling, zc, v, epsr, lossTan, TFB):
    alpha = np.pi*np.sqrt(epsr)*Band.f/v*lossTan # attenuation constant in Np/m
    
    # calculate response of BB2
    BB2 = epsilon2*np.conj(TFB)*TFB*np.sinc(1j*alpha*lCoupling)*np.exp(-2.0*alpha*zc)
    return BB2

'''
Function to calculate the response of individual channels of arbitrary filter
bank using the "adjacent subtraction method."
'''
def ChannelResponse(Band, Data, lback, ZT, Z0=50.0, n=1.0, physSep = 0.25, approachFB=3, approachT=1):
    v = c/n # wavespeed    
    # create nChannels x f array to store the response of all channels
    Response = np.empty([np.shape(Data)[1], Band.npoints], dtype=complex)
    
    # array to store aggregate response of channels up to current-1
    PreviousResponse = np.zeros(Band.npoints, dtype = complex)
    for i in range(np.shape(Data)[1]):
        # data for FB up to and including channel i and line after channel i
        DataBefore = Data[:, 0:i+1]
         
        if i<np.shape(Data)[1]-1:
            # create transmission line after channel i
            lineLength = physSep*v/DataBefore[0, i]
            TLine = TransmissionLine(Band, lineLength, Z0, n)
            # FB up to and including channel i and line after channel i
            FBBefore = rf.connect(FilterBank(Band, DataBefore, Z0, n, physSep, approachFB), 1, TLine, 0)            
            
            # create FB of the remaining channels
            DataAfter = Data[:, i+1:np.shape(Data)[1]]
            FBAfter = TerminatedFilterBank(Band, DataAfter, lback, ZT, Z0, n, physSep, approachFB)
            
            # input impedance looking toward load at port 1 of FBAfter
            ZinAfter = Z0*(1.0+FBAfter.s[:,0,0])/(1.0-FBAfter.s[:,0,0])
        else:
            # transmission line after final channel before termination
            TLine = TransmissionLine(Band, lback, Z0, n)
            FBBefore = rf.connect(FilterBank(Band, DataBefore, Z0, n, physSep, approachFB), 1, TLine, 0)
            ZinAfter = ZT*np.ones(Band.npoints)
        
        # port reference impedances for FBBefore
        Z01 = Z0*np.ones(Band.npoints)
        Z02 = ZinAfter
        
        # convert to S-parameters with port 2 referenced to ZinAfter
        S11 = Z2S(FBBefore.z, Z01, Z02)[:,0,0]
        S21 = Z2S(FBBefore.z, Z01, Z02)[:,1,0]
        
        # aggregate response of channels up to and including channel i
        CurrentResponse = 1.0-np.conj(S11)*S11-np.conj(S21)*S21
        
        # calculate channel i response and store in Response
        CurChannelResponse = CurrentResponse-PreviousResponse
        Response[i,:]=CurChannelResponse
        
        PreviousResponse = CurrentResponse
    
    return Response
    
'''
Function to calculate the channel center frequencies of a band.

Arguments: lowest frequency, highest frequency, spectral resolution, 
oversampling ratio
Returns: # of channels, array of channel frequencies
'''
def ChannelFrequencies(fl, fu, R, Sigma):
    Nc = np.int(np.floor(Sigma*R*np.log(fu/fl)))  # number of channels, rounded down
    # print(Nc)
    x = np.exp(-(np.log(fu)-np.log(fl))/(Nc-1)) # frequency scaling of channels
    Channels = np.ones(Nc) # array to hold all the spectral channel frequencies
    Channels[0] = fu # initialize the first channel    
    for i in range(1, Nc):
        Channels[i]=x*Channels[i-1]
    return Nc, Channels
    
'''
Function to calculate S11 and S21 of an isolated channel.

Arguments: freq. band, coupling Q, internal Q
Returns: S11, S21 as a tuple.
'''
def IsolatedChannel(Band, fres, Qc, Qi):
    Qr = 1.0/(1.0/Qi+1.0/Qc)
    x = (Band.f-fres)/fres
    S21 = 1.0-(Qr/Qc)/(1.0+2.0j*Qr*x)
    S11 = S21-1.0
    return np.array([S11, S21])

#makeing spectral line and Tline one function since they are never apart
def connect_spectral_and_Tline(Band, Z0, fres_nxt, Qc_nxt, Qdet_nxt, Qloss_nxt, approach,v,physSep,epsr):
    NextSC = SpectralChannelLossy(Band, Z0, fres_nxt, Qc_nxt, Qdet_nxt, Qloss_nxt, approach,normalize = False)      
    # create interconnecting transmission line
    lambda_current = v/fres_nxt
    # lambda_nxt = v/fres_nxt
    lineLength = physSep*lambda_current
    # lineLength = physSep*(lambda_current+lambda_nxt)/2.0
    TLine = TransmissionLineLossy(Band, lineLength, Z0, v, epsr, 1/Qloss_nxt) 
    # connect Tline to spectral channel
    Ntwk = rf.connect(TLine, 1, NextSC, 0)
    return Ntwk


#define functions for multiprocessing
net_dict_global = {}
def connect_two(net1,net2):
    #print(net_dict_global)
    N = net_dict_global['net0'].nports
    net = rf.connect(net_dict_global['net'+str(net1)],N-1,net_dict_global['net'+str(net2)],0)
    return net
def multi_run_wrapper(args):
    return connect_two(*args)    
        
# function for connecting up the filter bank using multiple processors
# conncts them up in a pyrmid type fashion
# i.e. connects the nearest neighbers in the filter bank the repeats until there is just
# one network left unfortunaly the most time intensive network connctions are still just
# handeled with one processor.
# visually it looks like this
# N    N    N    N    N    N    N    N    N    N    N    N    N    N    N    N    N    N  
# |____|    |____|    |____|    |____|    |____|    |____|    |____|    |____|    |____|  could use 9 processors
#    N         N         N         N         N         N         N         N         N    
#    |_________|         |_________|         |_________|         |_________|         |    use 4 processsors
#         N                   N                   N                   N              N
#         |___________________|                   |___________________|              |    use 2 processors
#                   N                                       N                        N
#                   |_______________________________________|                        |    use 1 processor
#                                       N                                            N
#                                       |____________________________________________|    use 1 processor
#                                                                 N

'''
def connect_net_dict_multi(net_dict,processors):
    print(processors)
    global net_dict_global #probably bad coding edicate but it works
    net_dict_global = net_dict 
    len_networks = 10 #intilize while condition to get into the loop
    while len_networks>1: #while there is more that one network to be stiched together

        tot_networks = len(net_dict_global.keys())
        #print("tot_netorks", tot_networks)
        #make the list of indicies to connect
        connect_list = []
        if np.mod(tot_networks,2) == 0: #if even connect them all
            for i in range(0,tot_networks,2):
                connect_list.append((i,i+1))
        else: #if odd leave a straggler
            for i in range(0,tot_networks-1,2):
                connect_list.append((i,i+1))
        print(connect_list)
        pool = Pool(processes=processors) #intialize multiprocesssing pool
        result = pool.map(multi_run_wrapper, connect_list) #do the multiprocessing connecting
        pool.close() #close open files
        # put the results into a dictionary
        new_dict = {}
        len_result = len(result)
        #print("len_result", len_result)
        for i in range(0,len_result):
            new_dict['net'+str(i)] = result[i]
        if np.mod(tot_networks,2) == 1:
            new_dict['net'+str(i+1)] = net_dict_global['net'+str(tot_networks -1)]
        net_dict_global = new_dict
        len_networks = len(new_dict.keys())
        #print("assem",len(new_dict.keys()))
    return result
'''

def connect_net_dict_multi(net_dict,processors):
    #print(processors)
    global net_dict_global #probably bad coding edicate but it works
    net_dict_global = net_dict 
    len_networks = 10 #intilize while condition to get into the loop
    while len_networks>1: #while there is more that one network to be stiched together

        tot_networks = len(net_dict_global.keys())
        #print("tot_netorks", tot_networks)
        #make the list of indicies to connect
        connect_list = []
        if np.mod(tot_networks,2) == 0: #if even connect them all
            for i in range(0,tot_networks,2):
                connect_list.append((i,i+1))
        else: #if odd leave a straggler
            for i in range(0,tot_networks-1,2):
                connect_list.append((i,i+1))
        #print(connect_list)
        #print(len(connect_list))
        #pool = Pool(processes=processors) #intialize multiprocesssing pool #currently not using multiple processors but still connecting in pyramid fashion
        new_dict = {}
        N = net_dict_global['net0'].nports
        for i in range(0,len(connect_list)):
            #print("fast")
            result =  rf.connect(net_dict_global['net'+str(connect_list[i][0])],N-1,net_dict_global['net'+str(connect_list[i][1])],0)
            new_dict['net'+str(i)] = result
        if np.mod(tot_networks,2) == 1:
            new_dict['net'+str(i+1)] = net_dict_global['net'+str(tot_networks -1)]
        #result = pool.map(multi_run_wrapper, connect_list) #do the multiprocessing connecting
        #pool.close() #close open files
        # put the results into a dictionary
        len_result = len(result)
        #print("len_result", len_result)
        net_dict_global = new_dict
        len_networks = len(new_dict.keys())
        #print("assem",len(new_dict.keys()))
    return net_dict_global




import matplotlib.pyplot as plt
import numpy as np
import matplotlib
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

# make custom colormap
# purple mountian_majesty
def purple_mountain_majesty(type = 'smooth',plot = False):
    my_cmap  = np.ones((256,4))
    c1 = matplotlib.colors.to_rgb("#7C5AA3")
    c2 = matplotlib.colors.to_rgb("#9577B7")
    c3 = matplotlib.colors.to_rgb("#A493BE")
    c4 = matplotlib.colors.to_rgb("#EACDB4")
    c5 = matplotlib.colors.to_rgb("#E4A780")
    c6 = matplotlib.colors.to_rgb("#E4734A")

    all_cs = np.vstack((c1,c2,c3,c4,c5,c6))
    my_cmap[0:43,0:3] = c1
    my_cmap[43:86,0:3] = c2
    my_cmap[86:129,0:3] = c3
    my_cmap[129:172,0:3] = c4
    my_cmap[172:215,0:3] = c5
    my_cmap[215:256,0:3] = c6

    new_cmap = ListedColormap(my_cmap)

    my_cmap_2 = np.ones((256,4))
    for j in range(0,all_cs.shape[0]-1):
        for i in range(0,3):
            if j < all_cs.shape[0]-2:
                my_cmap_2[j*(256//(all_cs.shape[0]-1)):(j+1)*(256//(all_cs.shape[0]-1)),i] = np.linspace(all_cs[j,i],all_cs[j+1,i],256//(all_cs.shape[0]-1))
            else:
                my_cmap_2[j*(256//(all_cs.shape[0]-1)):256,i] = np.linspace(all_cs[j,i],all_cs[j+1,i],256-j*(256//(all_cs.shape[0]-1)))

    new_cmap_2 = ListedColormap(my_cmap_2)

    if plot:
        plt.figure(1)
        plt.title("purple_mountain_majesty_sequential")
        plt.imshow(np.ones((100,100))*np.linspace(0,1,100),cmap = new_cmap)
        plt.figure(2)
        for i in range(0,all_cs.shape[0]):
            result = plt.hist(np.random.randn(200)+i*2,color = all_cs[i,:],histtype = 'stepfilled',label = "C"+str(i),zorder = all_cs.shape[0]*2-2*i,alpha = 0.9,bins=15)
            bin_centers = ((result[1]+np.roll(result[1],1))/2.)[1:len(result[1])]
            plt.plot(np.hstack((result[1],result[1][-1]+(result[1][-1]-result[1][-2]))),np.hstack((0,result[0],0)),ls='steps',color = 'k',zorder = all_cs.shape[0]*2-1-2*i)
        plt.legend()
        plt.title("Natural for histograms")
        plt.figure(3)
        plt.title("purple_mountain_majesty")
        plt.imshow(np.ones((100,100))*np.linspace(0,1,100),cmap = new_cmap_2)
        plt.show()

    if type == 'sequential':
        return new_cmap
    elif type ==list:
        return all_cs
    else:
        return new_cmap_2




import numpy as np
import time
import datetime

from submm.instruments import anritsu2 as an

anritsu = an.Anritsu()
anritsu.set_power(10.)
ts = time.time()

times = np.array(())
sleep_time =48.0

time.sleep(50)

low_freq = 180.
high_freq =199.75
step = 0.25
mm_freqs = np.arange(low_freq,high_freq,step)

time_dir = "/home/superspec/Documents/data/kidpy_multitone/march2018/lo_chop_times/"+datetime.datetime.fromtimestamp(ts).strftime("%Y_%m_%d_%H_%M")

start = time.time()

for freq in mm_freqs:
    print(freq)
    anritsu.set_frequency(freq*10**9/18.)
    times = np.append(times,(time.time()-start))
    anritsu.turn_outputOn()
    time.sleep(sleep_time)
    times = np.append(times,(time.time()-start))
    anritsu.turn_outputOff()
    time.sleep(sleep_time)

#print times
np.save(time_dir,times)



import numpy as np
import matplotlib.pyplot as plt
import pkg_resources



# find the csv file where everit is
# feel free to edit excel file then save as csv to change font
stream = pkg_resources.resource_stream(__name__, 'font.csv')
my_bitmap_font = np.genfromtxt(stream,delimiter = ",",filling_values=0)


# this is class to handel asigning a new index to new shapes and subsets
# basically I like to have a string to describe everything but cif format
# wants an index so I use this class to map my string names to a nubmer i
# index which increaments with every new asignment 
class index_dictionary():
    
    def __init__(self):
        self.id = {}
        self.index_offset = 0 #allow for incrementing to start intergers >0
        self.appendix_name = '' #when nesting many similar subsets a subset name

    def add_new_index(self,name):
        if self.appendix_name != '':
            self.id[name+'_'+self.appendix_name] = self.index_offset+len(self.id)+1
            return name+'_'+self.appendix_name
        else:
            self.id[name] = self.index_offset+len(self.id)+1
            return name

    def lookup(self,name):
        if self.appendix_name != '':
            return self.id[name+'_'+self.appendix_name]
        else:
            return self.id[name]

def parse_index_input(index,name):
    if not isinstance(index, int):
        if name == '':
            print("name required is supplying an index class")
            return
        else:
            #print(index)
            name = index.add_new_index(name)
            num_index = index.id[name]
            #print("test")
            #print(num_index)
    else:
        num_index = index
    return index,num_index
    

def start_subset(f,index,scale_num = 1,scale_denom = 1,subset_name = '' ): #default scale is 1/100 of a micron
    index, num_index = parse_index_input(index,subset_name)
    f.write('DS '+str(num_index)+' '+str(scale_num)+' '+str(scale_denom)+';\n') 
    if subset_name != '':
        f.write('9 '+subset_name+';\n') # name subset

def end_subset(f):
    f.write('DF;\n') # end subset

def end_file(f):
    f.write('E;\n')

def translate(f,index,x,y):
    f.write('C'+str(index)+' T'+str(int(x))+','+str(int(y))+';\n')

def rotate(f,index,rotation):
    #print('C'+str(index)+' R'+str(int(np.cos(rotation*np.pi/180)*1000))+' '+str(int(np.sin(rotation*np.pi/180)*1000))+';\n')
    f.write('C'+str(index)+' R'+str(int(np.cos(rotation*np.pi/180)*1000000000))+' '+str(int(np.sin(rotation*np.pi/180)*1000000000))+';\n') #need lots of precision 

def mirror_y(f,index):
    #print('C'+str(index)+' R'+str(int(np.cos(rotation*np.pi/180)*1000))+' '+str(int(np.sin(rotation*np.pi/180)*1000))+';\n')
    f.write('C'+str(index)+' MY'+';\n')

def mirror_x(f,index):
    #print('C'+str(index)+' R'+str(int(np.cos(rotation*np.pi/180)*1000))+' '+str(int(np.sin(rotation*np.pi/180)*1000))+';\n')
    f.write('C'+str(index)+' MX'+';\n')


def draw_hexagon(f,index,diameter,x= 0,y= 0,layer = 1,scale_num = 1,scale_denom = 1,name=''):
    index, num_index = parse_index_input(index,name)
    '''
       _____    f is the file
      /     \   index is index for cif file
     /___.___\  x,y are dot 
     \   D   /  D is the diameter
      \_____/   scale_num in the scale numerator
                scale_denom is the scale denominator
    '''
    deg = np.pi/180.
    f.write('DS '+str(num_index)+' ' +str(scale_num)+' '+str(scale_denom)+ ';\n')
    if name != '':
        f.write('9 ' + name + ' ;\n')
    f.write('L L'+str(layer)+'D0;;\n') #layer 5
    f.write('P '+str(diameter//2)+','+str(0)+' '
                +str(int(np.cos(60*deg)*diameter/2))+','+str(int(np.sin(60*deg)*diameter/2))+' '
                +str(-int(np.cos(60*deg)*diameter/2))+','+str(int(np.sin(60*deg)*diameter/2))+' '
                +str(-diameter//2)+','+str(0)+' '
                +str(-int(np.cos(60*deg)*diameter/2))+','+str(-int(np.sin(60*deg)*diameter/2))+' '
                +str(int(np.cos(60*deg)*diameter/2))+','+str(-int(np.sin(60*deg)*diameter/2))
                +';\n')
    f.write('DF;\n') # end subset

def draw_rectangle(f,index,length,width,rotation = 0,x= 0,y= 0,
                       layer = 1,scale_num = 1,scale_denom = 1,name='',
                       fillet_corners = False,which_corners = 'all',corner_radius = 10,num_of_points = 100):
    '''
        l
  NW ________ NE
    |        |
    |    *   | w   * x,y
    |________|
  SW          SE
    '''
    index, num_index = parse_index_input(index,name)

    f.write('DS '+str(num_index)+' ' +str(scale_num)+' '+str(scale_denom)+ ';\n')
    if name != '':
        f.write('9 ' + name + ' ;\n')
    f.write('L L'+str(layer)+'D0;;\n') #layer 5
              #B length width xpos ypos [rotation] ;
    if fillet_corners:
        # just do a polynomial now
        # rotation does not work with fillet corners
        poly_str = 'P'+' '
        if which_corners == 'all':
            which_corners = ['NW','NE','SE','SW']
        if 'NW' in which_corners:
            angles = np.linspace(np.pi,np.pi/2,num_of_points)
            for i in range(0,len(angles)):
                poly_str += str(int(corner_radius*np.cos(angles[i])-length/2+corner_radius))+','+str(int(corner_radius*np.sin(angles[i])+width/2-corner_radius))+' '
        else:
            poly_str += str(int(-length/2))+','+str(int(width/2))+' '
            
        if 'NE' in which_corners:
            angles = np.linspace(np.pi/2,0,num_of_points)
            for i in range(0,len(angles)):
                poly_str += str(int(corner_radius*np.cos(angles[i])+length/2-corner_radius))+','+str(int(corner_radius*np.sin(angles[i])+width/2-corner_radius))+' '
        else:
            poly_str += str(int(+length/2))+','+str(int(width/2))+' '

        if 'SE' in which_corners:
            angles = np.linspace(4*np.pi/2,3*np.pi/2,num_of_points)
            for i in range(0,len(angles)):
                poly_str += str(int(corner_radius*np.cos(angles[i])+length/2-corner_radius))+','+str(int(corner_radius*np.sin(angles[i])-width/2+corner_radius))+' '
        else:
            poly_str += str(int(+length/2))+','+str(int(-width/2))+' '

        if 'SW' in which_corners:
            angles = np.linspace(3*np.pi/2,2*np.pi/2,num_of_points)
            for i in range(0,len(angles)):
                poly_str += str(int(corner_radius*np.cos(angles[i])-length/2+corner_radius))+','+str(int(corner_radius*np.sin(angles[i])-width/2+corner_radius))+' '
        else:
            poly_str += str(int(-length/2))+','+str(int(-width/2))+' '

        #print(poly_str)
            

        f.write(poly_str+';\n')
    else:
        if rotation == 0:
            f.write('B '+str(length)+' '+str(width)+' '+str(x)+' '+str(y)+';\n')
        else:
            f.write('B '+str(length)+' '+str(width)+' '+str(x)+' '+str(y)+' '+str(int(np.cos(rotation*np.pi/180)*1000))+' '+str(int(np.sin(rotation*np.pi/180)*1000))+';\n')

            
    f.write('DF;\n') # end subset



def draw_wire(f,index,thickness,x,y,layer = 1,scale_num = 1,scale_denom = 1,name  =''):
    index, num_index = parse_index_input(index,name)
    f.write('DS '+str(num_index)+' ' +str(scale_num)+' '+str(scale_denom)+ ';\n')
    if name != '':
        f.write('9 ' + name + ' ;\n')
    f.write('L L'+str(layer)+'D0;;\n') #layer 5
    wire_str = 'W'+str(thickness)+' '
    for i in range(0,len(x)):
        wire_str += str(int(x[i]))+','+str(int(y[i]))+' '

    f.write(wire_str+';\n')
    f.write('DF;\n') # end subset
    
       

def draw_wire_circle(f,index,radius,thickness,layer = 1,scale_num = 1,scale_denom = 1,name  =''):
    index, num_index = parse_index_input(index,name)
    f.write('DS '+str(num_index)+' ' +str(scale_num)+' '+str(scale_denom)+ ';\n')
    if name != '':
        f.write('9 ' + name + ' ;\n')
    f.write('L L'+str(layer)+'D0;;\n') #layer 5
    wire_str = 'W'+str(thickness)+' '
    angles = np.linspace(-np.pi,np.pi,100)
    for i in range(0,len(angles)):
        wire_str += str(int(radius*np.sin(angles[i])))+','+str(int(radius*np.cos(angles[i])))+' '

    f.write(wire_str+';\n')
    f.write('DF;\n') # end subset

def draw_circle(f,index,radius,layer = 1,scale_num = 1,scale_denom = 1,name  =''):
    index, num_index = parse_index_input(index,name)
    f.write('DS '+str(num_index)+' ' +str(scale_num)+' '+str(scale_denom)+ ';\n')
    if name != '':
        f.write('9 ' + name + ' ;\n')
    f.write('L L'+str(layer)+'D0;;\n') #layer 5
    circ_str = 'P'+' '
    angles = np.linspace(-np.pi,np.pi,100)
    for i in range(0,len(angles)):
        circ_str += str(int(radius*np.cos(angles[i])))+','+str(int(radius*np.sin(angles[i])))+' '

    f.write(circ_str+';\n')
    f.write('DF;\n') # end subset

def draw_arc(f,index,radius,width,start_angle,end_angle,layer = 1,scale_num = 1,scale_denom = 1,name  ='',num_of_points = 100):
    index, num_index = parse_index_input(index,name)
    f.write('DS '+str(num_index)+' ' +str(scale_num)+' '+str(scale_denom)+ ';\n')
    if name != '':
        f.write('9 ' + name + ' ;\n')
    f.write('L L'+str(layer)+'D0;;\n') #layer 5
    circ_str = 'P'+' '
    if start_angle>end_angle:
        #print("hello")
        num_of_points_1 = int(num_of_points*(360-start_angle)/(360-start_angle+end_angle))
        num_of_points_2 = num_of_points-num_of_points_1
        angles = np.hstack((np.linspace(np.pi*start_angle/180.,np.pi*359.99999/180,num_of_points_1),np.linspace(0,np.pi*end_angle/180,num_of_points_2)))
    else:
        angles = np.linspace(np.pi*start_angle/180.,np.pi*end_angle/180,num_of_points)
    for i in range(0,len(angles)):
        circ_str += str(int(radius*np.cos(angles[i])))+','+str(int(radius*np.sin(angles[i])))+' '
    for i in reversed(range(0,len(angles))):
        circ_str += str(int((radius+width)*np.cos(angles[i])))+','+str(int((radius+width)*np.sin(angles[i])))+' '

    f.write(circ_str+';\n')
    f.write('DF;\n') # end subset

def draw_polygon(f,index,x_points,y_points,layer = 1,scale_num = 1,scale_denom = 1,name  =''):
    index, num_index = parse_index_input(index,name)
    f.write('DS '+str(num_index)+' ' +str(scale_num)+' '+str(scale_denom)+ ';\n')
    if name != '':
        f.write('9 ' + name + ' ;\n')
    f.write('L L'+str(layer)+'D0;;\n') #layer 5
    poly_str = 'P'+' '
    for i in range(0,len(x_points)):
        poly_str += str(int(x_points[i]))+','+str(int(y_points[i]))+' '

    f.write(poly_str+';\n')
    f.write('DF;\n') # end subset

def draw_arc_spiral(f,index,radius,radius_2,width,start_angle,end_angle,layer = 1,scale_num = 1,scale_denom = 1,name  ='',num_of_points = 100):
    index, num_index = parse_index_input(index,name)
    f.write('DS '+str(num_index)+' ' +str(scale_num)+' '+str(scale_denom)+ ';\n')
    if name != '':
        f.write('9 ' + name + ' ;\n')
    f.write('L L'+str(layer)+'D0;;\n') #layer 5
    circ_str = 'P'+' '
    radii = np.linspace(radius,radius_2,num_of_points)
    if start_angle>end_angle:
        #print("hello")
        num_of_points_1 = int(num_of_points*(360-start_angle)/(360-start_angle+end_angle))
        num_of_points_2 = num_of_points-num_of_points_1
        angles = np.hstack((np.linspace(np.pi*start_angle/180.,np.pi*359.99999/180,num_of_points_1),np.linspace(0,np.pi*end_angle/180,num_of_points_2)))
    else:
        angles = np.linspace(np.pi*start_angle/180.,np.pi*end_angle/180,num_of_points)
    for i in range(0,len(angles)):
        circ_str += str(int(radii[i]*np.cos(angles[i])))+','+str(int(radii[i]*np.sin(angles[i])))+' '
    for i in reversed(range(0,len(angles))):
        circ_str += str(int((radii[i]+width)*np.cos(angles[i])))+','+str(int((radii[i]+width)*np.sin(angles[i])))+' '

    f.write(circ_str+';\n')
    f.write('DF;\n') # end subset

def draw_text(f,index,text,pixel_size = 1000,layer = 1,scale_num = 1,scale_denom = 1,name  =''):
    index, num_index = parse_index_input(index,name)
    # find new line characters
    text_lines = []
    for i in range(0,text.count('\n')+1):
        text_lines.append(text.split('\n')[i])
    f.write('DS '+str(num_index)+' ' +str(scale_num)+' '+str(scale_denom)+ ';\n')
    if name != '':
        f.write('9 ' + name + ' ;\n')
    f.write('L L'+str(layer)+'D0;;\n') #layer 5

    new_line = 0*pixel_size
    for text in text_lines:
        for i in range(0,len(text)):
            char_num = ord(text[i])
            if char_num < 48: #symbols 1
                char_num = ord(text[i])-32
                bitmap = my_bitmap_font[24:32,char_num*6:char_num*6+6]
                cap = 1
            elif char_num<58:#number
                char_num = ord(text[i])-48
                bitmap = my_bitmap_font[16:24,char_num*6:char_num*6+6]
                cap = 1
            elif char_num<65:#symbols 2
                char_num = ord(text[i])-58
                bitmap = my_bitmap_font[32:40,char_num*6:char_num*6+6]
                cap = 1
            elif char_num<91: # upper case
                char_num = ord(text[i])-65
                bitmap = my_bitmap_font[0:8,char_num*6:char_num*6+6]
                cap = 1
            elif char_num<97:#symbols 3
                char_num = ord(text[i])-91
                bitmap = my_bitmap_font[40:48,char_num*6:char_num*6+6]
                cap = 1
            elif char_num<123:#lower case
                char_num = ord(text[i])-97
                bitmap = my_bitmap_font[8:16,char_num*6:char_num*6+6]
                cap = 0
            else: #symbols 4
                char_num = ord(text[i])-123
                bitmap = my_bitmap_font[48:56,char_num*6:char_num*6+6]
                cap = 0


            #print(text[i])
            #print(bitmap)
            for j in range(0,8):
                for k in range(0,6):
                    if bitmap[j,k] != 0:
                        f.write('B '+str(pixel_size)+' '+str(pixel_size)+' '+str(pixel_size*k+i*pixel_size*6)+' '+str(8*pixel_size-pixel_size*j+cap*pixel_size-new_line)+';\n')
                    
        new_line = new_line+12*pixel_size

    f.write('DF;\n') # end subset
        





import os
import getpass
from typing import Union
from multiprocessing import Pool

import numpy as np
import matplotlib as mpl
from submm.KIDs.res.data_io import Fit
from submm.KIDs.res.fitting import fit_nonlinear_iq, fit_so_resonator_cable, guess_so_resonator_cable, ResonatorCable

# Debug mode
debug_mode = False
# multiprocessing
# the 'assumption' of max threads is that we are cpu limited in processing,
# so we use should not use more than a computer's available threads
max_threads = int(os.cpu_count())
# Try to strike a balance between the best performance and computer usability during processing
balanced_threads = max(max_threads - 2, 2)
# Use onl half of the available threads for processing
half_threads = int(np.round(os.cpu_count() * 0.5))
current_user = getpass.getuser()
if debug_mode:
    # this will do standard linear processing.
    multiprocessing_threads_default = None
    mpl.use(backend='module://backend_interagg')
elif half_threads < 2:
    multiprocessing_threads_default = None
else:
    multiprocessing_threads_default = half_threads



allowed_fits = ['fit_nonlinear_iq', 'fit_so_resonator_cable']
allowed_fits_int = set(range(len(allowed_fits)))
allowed_functions = {'fit_nonlinear_iq': fit_nonlinear_iq, 'fit_so_resonator_cable': fit_so_resonator_cable}


def fit_nonlinear_iq_wrapped(f_hz, z, tau=None, verbose: bool = True):
    if debug_mode:
        fit_single_res = fit_nonlinear_iq(f_hz, z, tau=tau, verbose=False)
        if verbose:
            fit_single_res.console()
        return fit_single_res
    try:
        fit_single_res = fit_nonlinear_iq(f_hz, z, tau=tau, verbose=False)
    except Exception as excep:
        print(repr(excep))
        print(f"failed to fit freq range: {np.min(f_hz) * 1e-6} - {np.max(f_hz) * 1e-6} MHz\n")
        return None
    else:
        if verbose:
            fit_single_res.console()
        return fit_single_res


def null_func(f_list, *args):
    nan_array = np.empty(len(f_list))
    nan_array[:] = np.nan
    return nan_array


def make_null_fit(f_hz, z, failed_str=None, flag_str=None):
    real = np.real(z)
    imag = np.imag(z)
    x0, bounds = guess_so_resonator_cable(f_hz, real, imag)
    guess = ResonatorCable(*x0)
    popt = np.empty(len(x0))
    popt[:] = np.nan
    pcov = np.empty((len(x0), len(x0)))
    pcov[:] = np.nan

    null_fit = Fit(origin=f'Any Failed Fit: {failed_str}',
                   func=null_func,
                   guess=guess, result=ResonatorCable(fr=(f_hz[0] + f_hz[-1]) / 2.0),
                   popt=popt, pcov=pcov,
                   f_data=f_hz, z_data=z, mask=None,
                   flags={flag_str})
    return null_fit


def fit_so_resonator_cable_wrapped(f_hz, z, verbose: bool = True):
    if debug_mode:
        fit_single_res = fit_so_resonator_cable(f_hz, z, verbose=False)
        if verbose:
            fit_single_res.console()
        return fit_single_res
    try:
        fit_single_res = fit_so_resonator_cable(f_hz, z, verbose=False)
    except RuntimeError as excep:
        excep_str = repr(excep)
        print(excep_str)
        print(f"\nfailed to fit freq range: {np.min(f_hz) * 1e-6} - {np.max(f_hz) * 1e-6} MHz")
        null_fit = make_null_fit(f_hz=f_hz, z=z, failed_str=excep_str, flag_str='failed-fit-runtimeerror')
        return null_fit
    else:
        if verbose:
            fit_single_res.console()
        return fit_single_res


wrapped_functions = {'fit_nonlinear_iq': fit_nonlinear_iq_wrapped,
                     'fit_so_resonator_cable': fit_so_resonator_cable_wrapped}


def fit_select(function_select: Union[str, int]):
    if isinstance(function_select, str):
        function_select = function_select.lower().strip().replace(' ', '_')
        for allowed_fit in allowed_fits:
            if function_select in allowed_fit:
                function_select = allowed_fit
                break
        else:
            raise ValueError(f"function_select must be one of {allowed_fits}")
        function_select = function_select
    elif isinstance(function_select, int):
        if function_select in allowed_fits_int:
            function_select = allowed_fits[function_select]
        else:
            raise ValueError(f"function_select must be one of {allowed_fits_int}")

    else:
        raise TypeError(f"function_select must be a string or int, not {type(function_select)}")
    f_wrapped = wrapped_functions[function_select]
    return f_wrapped


def fit_pool(f_hz_list, z_list, *args, fit_selection: Union[str, int] = 0,
             multiprocessing_threads: Union[int, None] = multiprocessing_threads_default,
             verbose: bool = True) -> list:
    """ For Handling N fits at once using multiprocessing. Elements of f_hz_list and z_list, must be the same length
    per item, but not between items.

    """
    f_wrapped = fit_select(fit_selection)
    if multiprocessing_threads is None:
        res_fits = []
        for f_hz, z in zip(f_hz_list, z_list):
            fit_single_res = f_wrapped(f_hz, z, *args, verbose)
            res_fits.append(fit_single_res)
    else:
        list_len = len(f_hz_list)
        arg_lists = [f_hz_list, z_list]
        for arg in args:
            arg_lists.append([arg] * list_len)
        arg_lists.append([verbose] * list_len)
        star_args = zip(*arg_lists)
        with Pool(multiprocessing_threads) as p:
            res_fits = [fit_single_res for fit_single_res in p.starmap(f_wrapped, star_args)]
    return res_fits


def fit_nonlinear_iq_pool(f_hz_list, z_list, tau: float = None, verbose: bool = True,
                          multiprocessing_threads: Union[int, None] = multiprocessing_threads_default) -> list:
    return fit_pool(f_hz_list, z_list, tau, fit_selection='fit_nonlinear_iq',
                    multiprocessing_threads=multiprocessing_threads, verbose=verbose)


def fit_so_resonator_cable_pool(f_hz_list, z_list, verbose: bool = True,
                                multiprocessing_threads: Union[int, None] = multiprocessing_threads_default) -> list:
    return fit_pool(f_hz_list, z_list, fit_selection='fit_so_resonator_cable',
                    multiprocessing_threads=multiprocessing_threads, verbose=verbose)






import numpy as np
import os
import pygetdata as gd
import struct
import matplotlib.pyplot as plt

def openStoredSweep(savepath,load_std = False):
    """Opens sweep data
       inputs:
           char savepath: The absolute path where sweep data is saved
       ouputs:
           numpy array Is: The I values
           numpy array Qs: The Q values"""
    files = sorted(os.listdir(savepath))
    I_list, Q_list, stdI_list, stdQ_list = [], [], [], []
    for filename in files:
        if filename.startswith('I'):
            I_list.append(os.path.join(savepath, filename))
        if filename.startswith('Q'):
            Q_list.append(os.path.join(savepath, filename))
        if filename.startswith('stdI'):
            stdI_list.append(os.path.join(savepath, filename))
        if filename.startswith('stdQ'):
            stdQ_list.append(os.path.join(savepath, filename))
    Is = np.array([np.load(filename) for filename in I_list])
    Qs = np.array([np.load(filename) for filename in Q_list])
    if len(stdI_list) >0:
            std_Is = np.array([np.load(filename) for filename in stdI_list])
            std_Qs = np.array([np.load(filename) for filename in stdQ_list])
    if load_std:
        return Is, Qs, std_Is, std_Qs
    else:
        return Is, Qs


# reads in an iq sweep and stors i and q and the frequencies in a dictionary
def read_iq_sweep(filename,load_std = False):
    if load_std:
        I, Q, I_std, Q_std = openStoredSweep(filename,load_std = True)
    else:
        I, Q, = openStoredSweep(filename)
    sweep_freqs = np.load(filename + '/sweep_freqs.npy')
    bb_freqs = np.load(filename + '/bb_freqs.npy')
    channels = len(bb_freqs)
    mags = np.zeros((channels, len(sweep_freqs))) 
    chan_freqs = np.zeros((len(sweep_freqs),channels))
    for chan in range(channels):
        chan_freqs[:,chan] = (sweep_freqs + bb_freqs[chan])/1.0e6
    if load_std:
        dict = {'I': I, 'Q': Q, 'freqs': chan_freqs,'I_std':I_std,'Q_std':Q_std}
    else:
        dict = {'I': I, 'Q': Q, 'freqs': chan_freqs}
    return dict

def read_stream(filename):
    firstframe = 0
    firstsample = 0
    d = gd.dirfile(filename, gd.RDWR|gd.UNENCODED)
    #print "Number of frames in dirfile =", d.nframes
    nframes = d.nframes
    
    vectors = d.field_list()
    ifiles = [i for i in vectors if i[0] == "I"]
    qfiles = [q for q in vectors if q[0] == "Q"]
    ifiles.remove("INDEX")
    ivals = d.getdata(ifiles[0], gd.FLOAT32, first_frame = firstframe, first_sample = firstsample, num_frames = nframes)
    qvals = d.getdata(qfiles[0], gd.FLOAT32, first_frame = firstframe, first_sample = firstsample, num_frames = nframes)
    ivals = ivals[~np.isnan(ivals)]
    qvals = qvals[~np.isnan(qvals)]
    i_stream = np.zeros((len(ivals),len(ifiles)))
    q_stream = np.zeros((len(qvals),len(qfiles)))
    
    for n in range(len(ifiles)):
        ivals = d.getdata(ifiles[n], gd.FLOAT32, first_frame = firstframe, first_sample = firstsample, num_frames = nframes)
        qvals = d.getdata(qfiles[n], gd.FLOAT32, first_frame = firstframe, first_sample = firstsample, num_frames = nframes)
        i_stream[:,n] = ivals[~np.isnan(ivals)]
        q_stream[:,n] = qvals[~np.isnan(qvals)]
    d.close()
    #read in the time file
    with open(filename+"/time",'rb') as content_file:
        content = content_file.read()
    time_val = []
    for i in range(0,len(content)/8):
        time_val.append(struct.unpack('d',content[0+8*i:8+8*i])[0])


    #read in the time file
    with open(filename+"/packet_count",'rb') as content_file:
        content = content_file.read()
    packet_val = []
    for i in range(0,len(content)/8):
        packet_val.append(struct.unpack('L',content[0+8*i:8+8*i])[0])
    packet = np.asarray(packet_val)
    if ((packet -np.roll(packet,1))[1:]!=1).any():#you dropped packet
        print("!!!!WARNING!!!!! you dropped some packets during your measurement consider increasing your system buffer size")
        plt.figure(1)
        plt.title("Delta t between packets")
        plt.plot((time_val-np.roll(time_val,1))[1:])
        plt.figure(2)
        plt.title("Delta packet")
        plt.plot((packet-np.roll(packet,1))[1:])
        plt.show()
    
    dictionary = {'I_stream':i_stream,'Q_stream':q_stream,'time':time_val,'packet_count':packet_val}
    return dictionary
	
